import java.io.*;
import java.util.*;

public class CountCompare_Exact {
	
	// general information
    public static SpliceGraph s;
	public static Edge[] edge_list;
	public static String[] reads;
	public static ArrayList[] align;
	public static Integer[][] pos;
	public static Random rand;
	public static int fasta_readlen;
	
	// information for each bootstrap run
	public static String[] reads_1;
	public static ArrayList[] align_1;
	
    public static void main(String[] args) {
		if (args.length != 3) {
			System.out.println("Usage: CountCompare_Bowtie <psg> <readfile> <numTrial>");
			System.exit(0);
		}
		
		String psg = args[0];
		s = new SpliceGraph(psg);
		double[] initial = s.getEdgeWeights();
		int verbose = 0;
		
		//int howLong = Integer.parseInt(args[verbose+1]);
		//int numTrial = Integer.parseInt(args[verbose+2]);
		
		String readfile	= args[1];
		process_reads(readfile);
		int numTrial = Integer.parseInt(args[2]);
		align = s.readSubPaths(reads);
		
		// checking the edge case where we don't have any aligned reads
		if (reads.length == 0) {
			double ll = 0.0;
			//System.out.println("\tNumber of null reads: 0");
			System.out.println("\tNumber of reads: 0");
			s.emOutput(ll,-1);
			
			System.out.println();
			System.out.println("Current log likelihood: "+ll);
			System.out.println("Converged after 0 iterations");
			System.out.println("Log likelihood diff: 0");
			System.out.println("Max edge difference: 0.0");
			System.out.println("L1 norm: 0.0");
			System.out.println("L2 norm: 0.0");
			return;
		}
		
		// so if we're here, then we definitely have reads
		rand = new Random();
		
		double[][][] read_learned = new double[4][numTrial][initial.length];
		double[][][] count_learned = new double[4][numTrial][initial.length];
		
		for (int i=0; i<numTrial; i++) {
			s.resetEdgeWeights();
			generateReads(10);
			s.learnEM(reads_1,align_1,verbose);
			read_learned[0][i] = s.getEdgeWeights();
			s.resetEdgeWeights();
			JunctionRead.edgecount(align_1,s);
			count_learned[0][i] = s.getEdgeWeights();
			
			s.resetEdgeWeights();
			generateReads(100);
			s.learnEM(reads_1,align_1,verbose);
			read_learned[1][i] = s.getEdgeWeights();
			s.resetEdgeWeights();
			JunctionRead.edgecount(align_1,s);
			count_learned[1][i] = s.getEdgeWeights();
			
			s.resetEdgeWeights();
			generateReads(1000);
			s.learnEM(reads_1,align_1,verbose);
			read_learned[2][i] = s.getEdgeWeights();
			s.resetEdgeWeights();
			JunctionRead.edgecount(align_1,s);
			count_learned[2][i] = s.getEdgeWeights();
			
			s.resetEdgeWeights();
			generateReads(10000);
			s.learnEM(reads_1,align_1,verbose);
			read_learned[3][i] = s.getEdgeWeights();
			s.resetEdgeWeights();
			JunctionRead.edgecount(align_1,s);
			count_learned[3][i] = s.getEdgeWeights();
			
			/*s.resetEdgeWeights();
			generateReads(100000);
			s.learnEM(reads_1,align_1,verbose);
			read_learned[4][i] = s.getEdgeWeights();
			s.resetEdgeWeights();
			JunctionRead.edgecount(align_1,s);
			count_learned[4][i] = s.getEdgeWeights();*/
		}
		System.out.println();
		
		calculate_variance(read_learned,initial);
		calculate_variance(count_learned,initial);
    }
	
	// prints out average variance over weights
	private static void calculate_variance(double[][][] weights, double[] initial) {
		// step 1: find mean weight for each edge
		double[][] means = new double[weights.length][weights[0][0].length];
		for (int i=0; i<weights.length; i++) {
			for (int j=0; j<weights[i][0].length; j++) {
				for (int k=0; k<weights[i].length; k++) {
					means[i][j] += weights[i][k][j];
				}
				means[i][j] /= (double) weights[i].length;
			}
		}
		// step 2: find the average squared difference from the mean
		double[][] variances = new double[weights.length][weights[0][0].length];
		for (int i=0; i<weights.length; i++) {
			for (int j=0; j<weights[i][0].length; j++) {
				for (int k=0; k<weights[i].length; k++) {
					variances[i][j] += Math.pow(means[i][j]-weights[i][k][j],2);
				}
				variances[i][j] /= (double) weights[i].length;
				// EDIT: now returns standard deviation
				variances[i][j] = Math.sqrt(variances[i][j]);
			}
		}
		// step 3: print results
		System.out.println("Means:");
		edge_list = s.getEdgeList();
		for (int i=0; i<means[0].length; i++) {
			if (!s.is_5prime(edge_list[i])) {
				for (int j=0; j<means.length; j++) {
					System.out.printf("\t%.3f",means[j][i]);
				}
				System.out.printf("\t%.3f",initial[i]);
				System.out.println();
			}
		}
		System.out.println("Variances:");
		for (int i=0; i<variances[0].length; i++) {
			if (!s.is_5prime(edge_list[i])) {
				for (int j=0; j<variances.length; j++) {
					System.out.printf("\t%.3f",variances[j][i]);
				}
				System.out.println();
			}
		}
	}
	
	// chooses specified number of random reads with replacement
	private static void generateReads(int howMany) {
		reads_1 = new String[howMany];
		align_1 = new ArrayList[howMany];
		for (int i=0; i<howMany; i++) {
			int which = rand.nextInt(reads.length);
			reads_1[i] = reads[which];
			align_1[i] = align[which];
		}
	}
	
	// reads information from a bowtie alignment file
	private static void process_reads(String filename) {
		ArrayList<String> a = new ArrayList<String>();
		
		try {
			Scanner s = new Scanner(new File(filename));
			while (s.hasNextLine()) {
				a.add(s.nextLine());
			}
		}
		catch (IOException ioe) {
			System.out.println("Error reading "+filename);
			System.exit(0);
		}
		reads = Utility.StringArray(a);
	}
	
	// retrieves the reads from the read arraylist
	private static String[] getReads(ArrayList<String> r, ArrayList<String> s) {
		int num = count_unique(r);
		String[] result = new String[num];
		String id = null;
		int index = 0;
		ArrayList<String> used_ids = new ArrayList<String>();
		for (int i=0; i<r.size(); i++) {
			if (!contains(used_ids, r.get(i))) {
				id = r.get(i);
				used_ids.add(id);
				result[index] = s.get(i);
				index++;
			}
		}
		return trim(result);
    }
	
	private static String[] trim(String[] to_trim) {
		int i=0;
		for (i=0; i<to_trim.length; i++) {
			if (to_trim[i] == null) {
				break;
			}
		}
		String[] result = new String[i];
		for (i=0; i<result.length; i++) {
			result[i] = to_trim[i];
		}
		return result;
    }
	
	// Assumption: any duplicates will be consecutive
    private static int count_unique(ArrayList<String> r) {
		if (r.size() == 0) { return 0; }
		int unique = 1;
		int index = 1;
		String current = r.get(0);
		while (index < r.size()) {
			if (!r.get(index).equals(current)) {
				current = r.get(index);
				unique++;
			}
			index++;
		}
		return unique;
    }
	
	private static boolean contains(ArrayList<String> a, String b) {
		boolean retval = false;
		for (int i=0; i<a.size(); i++) {
			if (a.get(i).equals(b)) {
				retval = true;
			}
		}
		return retval;
    }
    
    private static boolean contains(ArrayList<Integer[]> a, Integer[] b) {
		boolean retval = false;
		for (int i=0; i<a.size(); i++) {
			if (a.get(i).length == b.length) {
				boolean verify = true;
				for (int j=0; j<a.get(i).length; j++) {
					if (a.get(i)[j] != b[j]) {
						verify = false;
					}
				}
				if (verify) {
					retval = true;
					break;
				}
			}
		}
		return retval;
    }
}