import java.util.*;
import java.io.*;

public class SpliceGraph {
    // Data members
    private Segment[] segments;
    private Edge[] edge_list;
    private Edge[][] edges;
    private double[][] f;   // sum of weight over all paths i,j
    private double[] z;     // expected edge counts for edge i,j
    private Random rand;
    private int[][] connected;
	private double[] lambda;
	private double[] beta;
	
    private int FRAG_MEAN = 200;
    private double FRAG_STDEV = Math.sqrt(200);
	private double BETA_VALUE = 1.0;
    
    ///////////////////////////////////////////////////////////////////////////
    // Constructor Methods
    ///////////////////////////////////////////////////////////////////////////
    
    // Constructor
    public SpliceGraph (String filename) {
		int section_start = -1;
		int line_num = 0;
		String[] lines = getLines(filename);
		this.rand = new Random();
		
		while (line_num <= lines.length) {
			if (line_num == lines.length || lines[line_num].startsWith("[")) {
				if (section_start != -1) {
					String section = lines[section_start].trim();
					section = section.substring(1,section.length()-1);
					String[] s_lines = getSectionLines(lines,
													   section_start,line_num);
					if (section.equalsIgnoreCase("segments")) {
						readSegments(s_lines);
					}
					else if (section.equalsIgnoreCase("edges")) {
						// creates first order connect and reach matrices
						readEdges(s_lines);
					}
					else if (section.equalsIgnoreCase("sequence")) {
						System.out.println("PSG uses old format, please use "+
										   "PSGTranslator");
						System.exit(0);
					}
				}
				section_start = line_num;
			}
			line_num++;
		}
		
		lambda = new double[segments.length-1];
		beta = new double[edge_list.length];
		initialize_lambda();
		
		if (this.segments == null || this.edges == null || 
			this.edge_list == null) {
			System.out.print("Required section not found; check PSG format.");
			System.exit(0);
		}
    }
    
    // Copy constructor
    public SpliceGraph(SpliceGraph copy) {
		this.rand = new Random();
		
		// copy segments
		Segment[] s = copy.getSegments();
		this.segments = new Segment[s.length];
		for (int i=0; i<s.length; i++) {
			this.segments[i] = new Segment(s[i]);
		}
		
		// copy the edge list and first order connectivity matrix
		Edge[] el = copy.getEdgeList();
		this.edge_list = new Edge[el.length];
		this.edges = new Edge[s.length][s.length];
		for (int i=0; i<el.length; i++) {
			int source = copy.getSegIndex(el[i].source);
			int target = copy.getSegIndex(el[i].target);
			this.edge_list[i] = new Edge();
			this.edge_list[i].source = this.segments[source];
			this.edge_list[i].target = this.segments[target];
			this.edge_list[i].weight = el[i].weight;
			this.edges[source][target] = this.edge_list[i];
		}
		
		// copy the reachability matrix
		int[][] con = copy.getConnected();
		this.connected = new int[con.length][con[0].length];
		for (int i=0; i<con.length; i++) {
			for (int j=0; j<con[i].length; j++) {
				this.connected[i][j] = con[i][j];
			}
		}
    }
    
	/****************************************************************
	 // Copy constructor number TWO
	 // NOTE: I don't think this is ever used
	 public SpliceGraph(SpliceGraph copy, int[][] transcripts) {
	 this.rand = new Random();
	 this.segments = get_seglist(copy.getSegments(), transcripts);
	 int[][] t = map_trans(copy.getSegments(), transcripts);
	 
	 get_edgelist(t);
	 get_edges();
	 }
	 
	 // Gets the edge list for a transcriptome 
	 // NOTE: I don't think this is ever used
	 private void get_edgelist(int[][] trans) {
	 ArrayList<Edge> e = new ArrayList<Edge>();
	 for (int i=0; i<trans.length; i++) {
	 for (int j=0; j<trans[i].length-1; j++) {
	 int source = trans[i][j];
	 int target = trans[i][j+1];
	 Edge e1 = new Edge();
	 e1.source = this.segments[source];
	 e1.target = this.segments[target];
	 if (source == 0) {
	 e1.weight = 1/(double) trans.length;
	 }
	 else {
	 e1.weight = 1;
	 }
	 e.add(e1);
	 }
	 }
	 this.edge_list = Utility.EdgArray(e);
	 }
	 
	 // Gets the appropriate edge setup for a transcriptome
	 // NOTE: I don't think this is ever used
	 private void get_edges() {
	 this.edges = new Edge[this.segments.length][this.segments.length];
	 for (int i=0; i<this.edge_list.length; i++) {
	 Edge temp = this.edge_list[i];
	 int source = getSegIndex(temp.source);
	 int target = getSegIndex(temp.target);
	 this.edges[source][target] = temp;
	 }
	 }
	 
	 // maps the line transcripts to our new segment set
	 // NOTE: I don't think this is ever used
	 private int[][] map_trans(Segment[] copy, int[][] transcripts) {
	 // figure out how many of each we'll use
	 int[] num_segs = new int[copy.length];
	 num_segs[0] = 1;
	 for (int i=0; i<transcripts.length; i++) {
	 for (int j=0; j<transcripts[i].length; j++) {
	 num_segs[transcripts[i][j]]++;
	 }
	 }
	 int[] used_seg = new int[copy.length];
	 int[][] new_trans = new int[transcripts.length][];
	 for (int i=0; i<transcripts.length; i++) {
	 int l = transcripts[i].length+1;
	 if (transcripts[i][transcripts[i].length-1]!=copy.length-1) {
	 l++;
	 }
	 int[] t = new int[l];
	 for (int j=0; j<transcripts[i].length; j++) {
	 int num_prev = 0;
	 for (int k=0; k<transcripts[i][j]; k++) {
	 num_prev += num_segs[k];
	 }
	 t[j+1] = used_seg[transcripts[i][j]]+num_prev;
	 used_seg[transcripts[i][j]]++;
	 }
	 t[t.length-1] = this.segments.length-1;
	 new_trans[i] = t;
	 }
	 return new_trans;
	 }
	 
	 // Gets the appropriate segment setup for a transcriptome
	 // NOTE: I don't think this is ever used
	 private Segment[] get_seglist(Segment[] copy, int[][] transcripts) {
	 int[] num_segs = new int[copy.length];
	 int total_segs = 1;
	 for (int i=0; i<transcripts.length; i++) {
	 for (int j=0; j<transcripts[i].length; j++) {
	 num_segs[transcripts[i][j]]++;
	 total_segs++;
	 }
	 }
	 if (num_segs[num_segs.length-1] > 1) {
	 total_segs -= num_segs[num_segs.length-1]-1;
	 num_segs[num_segs.length-1] = 1;
	 }
	 else if (num_segs[num_segs.length-1] == 0) {
	 num_segs[num_segs.length-1]++;
	 total_segs++;
	 }
	 Segment[] s = new Segment[total_segs];
	 s[0] = new Segment(copy[0]);
	 int index = 1;
	 for (int i=0; i<num_segs.length; i++) {
	 for (int j=0; j<num_segs[i]; j++) {
	 s[index] = new Segment(copy[i]);
	 index++;
	 }
	 }
	 return s;
	 }
	 ****************************************************************/
    
    // Retrieves all data from file and stores it in a String array
    private String[] getLines(String filename) {
		ArrayList<String> lines = new ArrayList<String>();
		try {
			Scanner r = new Scanner(new File(filename));
			while (r.hasNextLine()) {
				// discards all comment lines
				String line = r.nextLine();
				line = trimData(line);
				if (line.length() > 0) {
					lines.add(line);
				}
			}
		}
		catch (IOException ioe) {
			System.out.println(filename+" contains an error.");
			System.exit(0);
		}
		return Utility.StringArray(lines);
    }
    
    // Verifies that a line contains data and not a comment
    // If comment, return value is null
    private String trimData(String s) {
		String r = s.trim();
		if (r.startsWith("#")) {
			r = "";
		}
		return r;
    }
    
    // Retrieves all data from a section and stores it in a String array
    private String[] getSectionLines(String[] a, int start, int line_num) {
		ArrayList<String> s_lines = new ArrayList<String>();
		for (int i=start+1; i<line_num; i++) {
			s_lines.add(a[i]);
		}
		return Utility.StringArray(s_lines);
    }
	
	/***************************************
	 // Reads sequence data
	 // THIS IS NEVER CALLED
	 private void readSequence(String[] a) {
	 this.sequence = "";
	 for (int i=0; i<a.length; i++) {
	 this.sequence += a[i];
	 }
	 }
	 ****************************************/
    
    // Reads segment data
    private void readSegments(String[] a) {
		ArrayList<Segment> segs = new ArrayList<Segment>();
		for (int i=0; i<a.length; i++) {
			Segment temp = new Segment();
			String[] s = a[i].split("[\\s]+");
			if (s.length > 1) {
				// segments 0 and M have no sequence data
				temp.sequence = s[1];
			}
			//temp.start = Integer.parseInt(s[1]);
			//temp.end = Integer.parseInt(s[2]);
			segs.add(temp);
		}
		this.segments = Utility.SegArray(segs);
    }
    
    // Reads edge data
    private void readEdges(String[] a) {
		ArrayList<Edge> eds = new ArrayList<Edge>();
		for (int i=0; i<a.length; i++) {
			Edge temp = new Edge();
			String[] s = a[i].split("[\\s]+");
			try {
				temp.source = this.segments[Integer.parseInt(s[0])];
				temp.target = this.segments[Integer.parseInt(s[1])];
			}
			catch (NullPointerException npe) {
				System.out.println("ERROR: PSG format requires segments"+
								   " to be listed before edges.");
				System.exit(0);
			}
			temp.weight = Double.parseDouble(s[2]);
			eds.add(temp);
		}
		makeAdjMatrix(eds);
		makeReachMatrix();
		this.edge_list = Utility.EdgArray(eds);
    }
    
    // Creates an adjacency matrix
    // if a value is null, there is no connecting edge
    private void makeAdjMatrix(ArrayList<Edge> eds) {
		this.edges = new Edge[this.segments.length][this.segments.length];
		for (int i=0; i<eds.size(); i++) {
			Edge temp = eds.get(i);
			int source = getSegIndex(temp.source);
			int target = getSegIndex(temp.target);
			edges[source][target] = temp;
		}
    }
	
	// Creates reachability matrix
	//  0: not reachable
	//  1: i is earlier in partial ordering than j (or i==j)
	// -1: j is earlier in partial ordering than i
	// Requires edges[][] to have been created and filled out
	private void makeReachMatrix() {
		this.connected = new int[this.segments.length][this.segments.length];
		// first order connectivity
		for (int i=0; i<this.edges.length; i++) {
			if (i == 0) {
				// everything is reachable from segment 0
				// segment 0 is before everything
				for (int j=0; j<this.edges[i].length; j++) {
					this.connected[j][i] = -1;
					this.connected[i][j] = 1;
				}
			}
			else {
				for (int j=i; j<this.edges[i].length; j++) {
					if (j == i) {
						this.connected[i][j] = 1;
					}
					else if (this.edges[i][j] != null) {
						// i is before j
						this.connected[i][j] = 1;
						this.connected[j][i] = -1;
					}
					else if (this.edges[j][i] != null) {
						// j is before i
						this.connected[j][i] = 1;
						this.connected[i][j] = -1;
					}
					if (j == this.edges[i].length-1) {
						// everything is reachable from segment M
						// segment M is after everything else
						this.connected[i][j] = 1;
						this.connected[j][i] = -1;
					}
				}
			}
		}
		
		// second pass - all other connectivity?
		for (int i=this.connected.length-2; i>0; i--) {
			for (int j=this.connected[i].length-2; j>i; j--) {
				if (this.connected[i][j] == 1) {
					// i comes before j
					// go to column [i] and see what else it's connected to
					for (int k=i-1; k>0; k--) {
						if (this.connected[k][i] == 1) {
							// k comes before i; therefore before j
							this.connected[k][j] = 1;
							this.connected[j][k] = -1;
						}
					}
				}
				else if (this.connected[i][j] == -1) {
					// i comes after j
					// go to column [i] and see what else it's connected to
					for (int k=i-1; k>0; k--) {
						if (this.connected[k][i] == -1) {
							// k comes after i, therefore after j
							this.connected[k][j] = -1;
							this.connected[j][k] = 1;
						}
					}
				}
			}
		}
	}
    
    // prints out the segment strings
    public void printSegments() {
		for (int i=0; i<segments.length; i++) {
			System.out.println(segments[i].sequence);
		}
    }
    
    ///////////////////////////////////////////////////////////////////////////
    // Methods for Generating Single Reads
    ///////////////////////////////////////////////////////////////////////////
    
    // Generates a number of reads of a specified length
    // All reads of equal length, and all reads exact (no read errors)
    public String[] generateReads(int howMany, int howLong) {
		String[] reads = new String[howMany];
		double d_alpha = 0;
		double[] thetas = new double[segments.length];
		
		calculate_f();
		
		for (int i=0; i<segments.length; i++) {
			d_alpha += f[0][i]*segments[i].length();
		}
		for (int i=0; i<segments.length; i++) {
			thetas[i] = (f[0][i]*segments[i].length())/d_alpha;
		}
		
		for (int i=0; i<howMany; i++) {
			reads[i] = sampleRead(howLong,thetas);
		}
		return reads;
    }
    
	// Generates reads, paired with their corresponding alignments
	// All reads of equal length, all reads exact (no read errors)
	public String[] generateReads_Aligned(int howMany, int howLong) {
		String[] reads = new String[howMany];
		double d_alpha = 0;
		double[] thetas = new double[segments.length];
		
		calculate_f();
		
		for (int i=0; i<segments.length; i++) {
			d_alpha += f[0][i]*segments[i].length();
		}
		for (int i=0; i<segments.length; i++) {
			thetas[i] = (f[0][i]*segments[i].length())/d_alpha;
		}
		
		for (int i=0; i<howMany; i++) {
			reads[i] = sampleRead_Aligned(howLong,thetas);
		}
		return reads;
	}
	
	// Sampling method, allowing overflow into poly-A tail
    private String sampleRead_Aligned(int howLong, double[] thetas) {
		int a = getStartSeg(thetas);
		Segment s = segments[a];
		int b = this.rand.nextInt(s.length()); // starting position
		String result = getReadSequence_Aligned(howLong,s,b,thetas);
		return result;
    }
	
    // Sampling method, allowing overflow into poly-A tail
    private String sampleRead(int howLong, double[] thetas) {
		int a = getStartSeg(thetas);
		Segment s = segments[a];
		int b = this.rand.nextInt(s.length()); // starting position
		String result = getReadSequence(howLong,s,b,thetas);
		return result;
    }
    
    // Randomly chooses our start segment according to the f values
    private int getStartSeg(double[] thetas) {
		double a = this.rand.nextDouble();
		double total = 0;
		for (int i=0; i<thetas.length; i++) {
			total += thetas[i];
			if (a < total) {
				return i;
			}
		}
		return -1;
    }
	
	// Returns the read sequence based on our segment and start position
    // Now featuring reads over three or more segments!
    private String getReadSequence_Aligned(int howLong, Segment s, int b,
										   double[] thetas) {
		String retval = "";
		String alig = ""+getSegIndex(s);
		//String s_bases = this.sequence.substring(s.start,s.end);
		String s_bases = s.sequence;
		if (b+howLong < s_bases.length()) {
			retval = s_bases.substring(b,b+howLong);
		}
		else {
			retval = s_bases.substring(b);
			howLong -= s_bases.length()-b;
			Segment t = s;
			while (howLong > 0) {
				t = getNextSeg(t);
				if (t != null) {
					if (t.length() > howLong) {
						// last pass
						//retval += this.sequence.substring(t.start,t.start+howLong);
						alig += ","+getSegIndex(t);
						retval += t.sequence.substring(0,howLong);
						howLong -= howLong;
					}
					else {
						//retval += this.sequence.substring(t.start,t.end);
						alig += ","+getSegIndex(t);
						retval += t.sequence;
						howLong -= t.length();
					}
				}
				else {
					// add a poly-A tail
					for (int i=0; i<howLong; i++) {
						retval += "A";
					}
					//alig += ","+(segments.length-1);
					howLong = 0;
				}
			}
		}
		retval += "\t"+alig;
		return retval;
    }
    
    // Returns the read sequence based on our segment and start position
    // Now featuring reads over three or more segments!
    private String getReadSequence(int howLong, Segment s, int b,
								   double[] thetas) {
		String retval = "";
		//String s_bases = this.sequence.substring(s.start,s.end);
		String s_bases = s.sequence;
		if (b+howLong < s_bases.length()) {
			retval = s_bases.substring(b,b+howLong);
		}
		else {
			retval = s_bases.substring(b);
			howLong -= s_bases.length()-b;
			Segment t = s;
			while (howLong > 0) {
				t = getNextSeg(t);
				if (t != null) {
					if (t.length() > howLong) {
						// last pass
						//retval += this.sequence.substring(t.start,t.start+howLong);
						retval += t.sequence.substring(0,howLong);
						howLong -= howLong;
					}
					else {
						//retval += this.sequence.substring(t.start,t.end);
						retval += t.sequence;
						howLong -= t.length();
					}
				}
				else {
					// add a poly-A tail
					for (int i=0; i<howLong; i++) {
						retval += "A";
					}
					howLong = 0;
				}
			}
		}
		return retval;
    }
    
    // chooses a subsequent segment based on our current position
    private Segment getNextSeg(Segment s) {
		int which = getSegIndex(s);
		Segment result = null;
		if (which != this.segments.length-1) {
			double a = this.rand.nextDouble();
			double total = 0;
			for (int i=0; i<edges[which].length; i++) {
				if (edges[which][i] != null) {
					total += edges[which][i].weight;
					if (a<total) {
						result = this.segments[i];
						break;
					}
				}
			}
		}
		return result;
    }
    
    ///////////////////////////////////////////////////////////////////////////
    // Methods for Generating Paired Reads
    ///////////////////////////////////////////////////////////////////////////
    
    // generates paired reads
    public String[][] generatePairedReads(int howMany, int howLong) {
		String[][] reads = new String[howMany][2];
		double d_alpha = 0;
		double[] thetas = new double[this.segments.length];
		
		calculate_f();
		
		for (int i=0; i<this.segments.length; i++) {
			d_alpha += f[0][i]*this.segments[i].length();
		}
		for (int i=0; i<this.segments.length; i++) {
			thetas[i] = (f[0][i]*this.segments[i].length())/d_alpha;
		}
		for (int i=0; i<howMany; i++) {
			int frLen = getFragLength(howLong);
			String frag = sampleRead(frLen,thetas);
			if (frLen > howLong) {
				reads[i][0] = frag.substring(0,howLong);
				reads[i][1] = frag.substring(frag.length()-howLong-1);
			}
			else {
				reads[i][0] = frag;
				reads[i][1] = frag;
			}
		}
		return reads;
    }
    
    // generates a fragment length from a Gaussian with mean 200 and var 200
	// (or whatever we set them at)
    private int getFragLength(int l) {
		double mean = FRAG_MEAN;
		double sd = FRAG_STDEV;
		double x = this.rand.nextGaussian();
		x *= sd;
		x += mean;
		return (int) Math.round(x);
    }
    
    ///////////////////////////////////////////////////////////////////////////
    // Hopcroft-Karp algorithm
    ///////////////////////////////////////////////////////////////////////////
    
    // globals that we'll need for this
    private int[] pairR;
    private int[] pairL;
    private int[] dist;
    
    // runs the HK algorithm
    public int[] hop_karp(int[][] comp) {
		pairL = new int[comp.length];
		pairR = new int[comp.length];
		dist = new int[comp.length+1];
		for (int i=0; i<comp.length; i++) {
			pairR[i] = -1;
			pairL[i] = -1;
		}
		int matching = 0;
		while (BFS(comp)) {
			for (int v=0; v<comp.length; v++) {
				if (pairL[v] == -1) {
					if (DFS(comp,v)) {
						matching++;
					}
				}
			}
		}
		return pairR;
    }
    
    // returns true if we have any connections
    // v is always an index to a node on the left
    private boolean DFS(int[][] comp, int v) {
		if (v != -1) {
			for (int u=0; u<comp[v].length; u++) {
				if (comp[v][u] == 1) {
					if (dist[pairR[u]+1] == dist[v+1]+1) {
						if (DFS(comp, pairR[u])) {
							pairR[u] = v;
							pairL[v] = u;
							return true;
						}
					}
				}
			}
			dist[v+1] = Integer.MAX_VALUE;
			return false;
		}
		return true;
    }
    
    // returns true if there is an unpaired edge
    private boolean BFS(int[][] comp) {
		Queue q = new Queue();
		for (int i=0; i<comp.length; i++) {
			if (pairL[i] == -1) {
				dist[i+1] = 0;
				q.enqueue(i);
			}
			else {
				dist[i+1] = Integer.MAX_VALUE;
			}
		}
		dist[0] = Integer.MAX_VALUE;
		while (!q.isEmpty()) {
			int v = q.dequeue();
			if (v != -1) {
				for (int u=0; u<comp[v].length; u++) {
					if (comp[v][u] == 1) {
						if (dist[pairR[u]+1] == Integer.MAX_VALUE) {
							dist[pairR[u]+1] = dist[v+1]+1;
							q.enqueue(pairR[u]);
						}
					}
				}
			}
		}
		return dist[0] != Integer.MAX_VALUE;
    }
    
    ///////////////////////////////////////////////////////////////////////////
    // Read Compatibility Methods
    ///////////////////////////////////////////////////////////////////////////
    
    // Get exon PSI values
    public double[] getPSI(String[] reads, ArrayList[] al) {
		double[] psi = new double[this.segments.length];
		double[] count = new double[this.segments.length];
		double total = 0;
		for (int i=0; i<reads.length; i++) {
			for (int j=0; j<al[i].size(); j++) {
				// should skip the unaligned reads
				Integer[] exons = (Integer[]) al[i].get(j);
				count[exons[0]]++;
				total++;
				if (exons.length > 1) {
					count[exons[1]]++;
					total++;
				}
			}
		}
		for (int i=0; i<psi.length; i++) {
			psi[i] = count[i]/reads.length;
		}
		return psi;
    }
	
	private void printMatrix(int[][] comp) {
		String filename = "matrix.csv";
		try {
			BufferedWriter out = new BufferedWriter(new FileWriter(filename));
			for (int i=0; i<comp.length; i++) {
				for (int j=0; j<comp[i].length; j++) {
					out.write(Integer.toString(comp[i][j]));
					if (j < comp[i].length-1) {
						out.write(",");
					}
				}
				out.newLine();
			}
			out.close();
		} catch (IOException ioe) {
			// do nothing
		}
    }
    
	/************************************************
	 // Finds the end position for the read alignment
	 // return value: absolute end position in sequence
	 private int getEndPos(String read, Integer[] seg) {
	 int index = -1;
	 boolean poly_a = true;
	 for (int i=0; i<read.length(); i++) {
	 if (read.charAt(i) != 'A') {
	 poly_a = false;
	 break;
	 }
	 }
	 if (poly_a) {
	 return this.sequence.length();
	 }
	 if (seg.length == 1) {
	 // single segment alignment
	 int start = this.segments[seg[0]].start;
	 int end = this.segments[seg[0]].end;
	 //String compare = this.sequence.substring(start,end);
	 String compare = this.segments[seg[0]].sequence;
	 index = compare.indexOf(read);
	 index += start;
	 index += read.length()-1;
	 }
	 else {
	 // gapped alignment
	 int start = this.segments[seg[0]].start;
	 int end = this.segments[seg[0]].end;
	 //String compare = this.sequence.substring(start,end);
	 String compare = this.segments[seg[0]].sequence;
	 int prevlen = 0;
	 int i=0;
	 do {
	 index = compare.indexOf(read);
	 if (index > -1) {
	 int endind = read.length()-(prevlen-index);
	 index = endind + this.segments[seg[i]].start;
	 break;
	 }
	 else {
	 i++;
	 start = this.segments[seg[i]].start;
	 end = this.segments[seg[i]].end;
	 prevlen = compare.length();
	 if (end-start > 0) {
	 //compare += this.sequence.substring(start,end);
	 compare += this.segments[seg[i]].sequence;
	 }
	 else {
	 for (int j=0; j<read.length()-1; j++) {
	 compare += "A";
	 }
	 }
	 //prevlen = compare.length();
	 }
	 } while (i<seg.length);
	 }
	 return index;
	 }
	 
	 // Finds the start position for the read alignment
	 // return value: absolute start position in sequence
	 private int getStartPos(String read, Integer[] seg) {
	 int index = -1;
	 boolean poly_a = true;
	 for (int i=0; i<read.length(); i++) {
	 if (read.charAt(i) != 'A') {
	 poly_a = false;
	 break;
	 }
	 }
	 if (poly_a) {
	 return this.sequence.length();
	 }
	 if (seg.length == 1) {
	 // single segment alignment
	 int start = this.segments[seg[0]].start;
	 int end = this.segments[seg[0]].end;
	 //String compare = this.sequence.substring(start,end);
	 String compare = this.segments[seg[0]].sequence;
	 index = compare.indexOf(read);
	 index += start;
	 }
	 else {
	 // gapped alignment
	 int start = this.segments[seg[0]].start;
	 int end = this.segments[seg[0]].end;
	 //String compare = this.sequence.substring(start,end);
	 String compare = this.segments[seg[0]].sequence;
	 int prevlen = 0;
	 for (int i=1; i<seg.length; i++) {
	 start = this.segments[seg[i]].start;
	 end = this.segments[seg[i]].end;
	 if (end-start > 0) {
	 //compare += this.sequence.substring(start,end);
	 compare += this.segments[seg[i]].sequence;
	 }
	 else {
	 for (int j=0; j<read.length()-1; j++) {
	 compare += "A";
	 }
	 }
	 index = compare.indexOf(read);
	 if (index > -1) {
	 int sum = 0;
	 for (int j=0; j<=i; j++) {
	 if (index < (sum+this.segments[seg[j]].length())) {
	 index -= sum;
	 index += this.segments[seg[j]].start;
	 break;
	 }
	 else {
	 sum += this.segments[seg[j]].length();
	 }
	 }
	 return index;
	 }
	 prevlen = compare.length();
	 }
	 }
	 return index;
	 }
	 *************************************************/
	
    // Checks whether read i and read j map to the same exons
    // Note: this method should be called for a line-graph structure
    // Multiple overlapping exons will cause this to suck
    private int[] sameExon(ArrayList<Integer[]> a, 
						   ArrayList<Integer[]> b) {
		int[] retval = {-1,-1};
		for (int i=0; i<a.size(); i++) {
			Integer[] x = a.get(i);
			for (int j=0; j<b.size(); j++) {
				Integer[] y = b.get(j);
				if (x[0].equals(y[0]) || x[x.length-1].equals(y[0]) || 
					(x.length > 1 && x[x.length-1].equals(y[y.length-1]) && 
					 !x[x.length-1].equals(this.segments.length-1))) {
						retval[0] = i;
						retval[1] = j;
						return retval;
					}
			}
		}
		return retval;
    }
    
    ///////////////////////////////////////////////////////////////////////////
    // Expectation Maximization Methods
    ///////////////////////////////////////////////////////////////////////////
    
    // Master method for paired-reads EM algorithm
    // Returns BIC value for this splice graph
    public double learnEM(String[][] reads, int v) {
		ArrayList[] sub_paths = readSubPaths(reads);
		
		// let's try just doing this once
		int[][] path_lengths = new int[sub_paths.length][];
		for (int i=0; i<sub_paths.length; i++) {
			path_lengths[i] = new int[sub_paths[i].size()];
			for (int j=0; j<sub_paths[i].size(); j++) {
				Integer[] path = (Integer[]) sub_paths[i].get(j);
				path_lengths[i][j] = getlen(reads[i],path);
			}
		}
		
		// convergence conditions
		double min_edge_up = .00001, min_inc_ll = .001;
		boolean done = false;
		
		// things to check
		double ll = eStep(sub_paths,path_lengths);
		int iter = 0;
		double wgt_update = 0;
		if (v==1) {
			emOutput(ll,iter-1);
		}
		
		float num_reads = 0;
		for (int i=0; i<this.edges[0].length; i++) {
			if (this.edges[0][i] != null) {
				int index = getEdgeIndex(this.edges[0][i]);
				num_reads += z[index];
			}
		}
		
		while (!done) {
			double prev_ll = ll;
			double[] prev_probs = new double[this.edge_list.length];
			for (int i=0; i<this.edge_list.length; i++) {
				prev_probs[i] = this.edge_list[i].weight;
			}
			
			//mStep();
			mStep(Math.round(num_reads));
			ll = eStep(sub_paths,path_lengths);
			wgt_update = getWeightUpdate(prev_probs);
			if (v == 1) {
				emOutput(ll,iter);
			}
			
			if (ll-prev_ll > min_inc_ll && wgt_update > min_edge_up) {
				iter++;
			}
			else {
				iter++;
				if (ll-prev_ll < 0 && Math.abs(ll-prev_ll) > min_inc_ll) {
					System.err.println("ERROR: decreasing likelihood");
					System.err.println("  Number of reads: "+reads.length);
					System.err.println("  Iterations: "+iter);
				}
				if (v == 1) {
					System.out.println();
					System.out.println("Current log likelihood: "+ll);
					System.out.println("Converged after "+iter+" iterations");
					System.out.println("Log likelihood diff: "+(ll-prev_ll));
					System.out.println("Max edge difference: "+wgt_update);
				}
				else {
				    //emOutput(ll,iter);
				}
				done = true;
			}
		}
		return bic(reads,ll);
    }
	
    // Master method for paired-reads EM algorithm
    // Returns BIC value for this splice graph
    public double learnEM(String[][] reads, ArrayList[] sub_paths, int v) {
		// let's try just doing this once
		int[][] path_lengths = new int[sub_paths.length][];
		for (int i=0; i<sub_paths.length; i++) {
			path_lengths[i] = new int[sub_paths[i].size()];
			for (int j=0; j<sub_paths[i].size(); j++) {
				Integer[] path = (Integer[]) sub_paths[i].get(j);
				path_lengths[i][j] = getlen(reads[i],path);
			}
		}
		
		// convergence conditions
		double min_edge_up = .00001, min_inc_ll = .001;
		boolean done = false;
		
		// things to check
		double ll = eStep(sub_paths,path_lengths);
		int iter = 0;
		double wgt_update = 0;
		if (v==1) {
			emOutput(ll,iter-1);
		}
		
		float num_reads = 0;
		for (int i=0; i<this.edges[0].length; i++) {
			if (this.edges[0][i] != null) {
				int index = getEdgeIndex(this.edges[0][i]);
				num_reads += z[index];
			}
		}
		
		while (!done) {
			double prev_ll = ll;
			double[] prev_probs = new double[this.edge_list.length];
			for (int i=0; i<this.edge_list.length; i++) {
				prev_probs[i] = this.edge_list[i].weight;
			}
			
			//mStep();
			mStep(Math.round(num_reads));
			ll = eStep(sub_paths,path_lengths);
			wgt_update = getWeightUpdate(prev_probs);
			if (v == 1) {
				emOutput(ll,iter);
			}
			
			if (ll-prev_ll > min_inc_ll && wgt_update > min_edge_up) {
				iter++;
			}
			else {
				iter++;
				if (ll-prev_ll < 0 && Math.abs(ll-prev_ll) > min_inc_ll) {
					System.err.println("ERROR: decreasing likelihood");
					System.err.println("  Number of reads: "+reads.length);
					System.err.println("  Iterations: "+iter);
				}
				if (v == 1) {
					System.out.println();
					System.out.println("Current log likelihood: "+ll);
					System.out.println("Converged after "+iter+" iterations");
					System.out.println("Log likelihood diff: "+(ll-prev_ll));
					System.out.println("Max edge difference: "+wgt_update);
				}
				else {
				    //emOutput(ll,iter);
				}
				done = true;
			}
		}
		return bic(reads,ll);
    }
	
    // Master method for paired-reads EM algorithm
    // Returns BIC value for this splice graph
    public double learnEM(String[][] reads, ArrayList[] sub_paths, 
						  ArrayList[] quality, int v, int[][] path_lengths) {
		// convergence conditions
		double min_edge_up = .00001, min_inc_ll = .001;
		boolean done = false;
		
		// things to check
		double ll = eStep(sub_paths,path_lengths,quality);
		int iter = 0;
		double wgt_update = 0;
		if (v==1) {
			emOutput(ll,iter-1);
		}
		
		float num_reads = 0;
		for (int i=0; i<this.edges[0].length; i++) {
			if (this.edges[0][i] != null) {
				int index = getEdgeIndex(this.edges[0][i]);
				num_reads += z[index];
			}
		}
		
		while (!done) {
			double prev_ll = ll;
			double[] prev_probs = new double[this.edge_list.length];
			for (int i=0; i<this.edge_list.length; i++) {
				prev_probs[i] = this.edge_list[i].weight;
			}
			//System.err.println("\tNumber of reads used: "+num_reads);
			
			//mStep();
			mStep(Math.round(num_reads));
			ll = eStep(sub_paths,path_lengths,quality);
			wgt_update = getWeightUpdate(prev_probs);
			if (v == 1) {
				emOutput(ll,iter);
			}
			
			if (ll-prev_ll > min_inc_ll && wgt_update > min_edge_up) {
				iter++;
			}
			else {
				iter++;
				if (ll-prev_ll < 0 && Math.abs(ll-prev_ll) > min_inc_ll) {
					System.err.println("ERROR: decreasing likelihood");
					System.err.println("  Number of reads: "+reads.length);
					System.err.println("  Iterations: "+iter);
				}
				if (v == 1) {
					System.out.println();
					System.out.println("Current log likelihood: "+ll);
					System.out.println("Converged after "+iter+" iterations");
					System.out.println("Log likelihood diff: "+(ll-prev_ll));
					System.out.println("Max edge difference: "+wgt_update);
				}
				else {
				    //emOutput(ll,iter);
				}
				done = true;
			}
		}
		return bic(reads,ll);
    }
	
    // Master method for running the EM algorithm
    // Returns the Bayesian Information Criterion value for this splice graph
    public double learnEM(String[] reads, int v) {
		ArrayList[] sub_paths = readSubPaths(reads);
		
		// convergence conditions
		double min_edge_up = .00001, min_inc_ll = .001;
		boolean done = false;
		
		// things to check
		double ll = eStep(sub_paths);
		int iter = 0;
		double wgt_update = 0;
		if (v == 1) { emOutput(ll,iter-1); }
		
		float num_reads = 0;
		for (int i=0; i<this.edges[0].length; i++) {
			if (this.edges[0][i] != null) {
				int index = getEdgeIndex(this.edges[0][i]);
				num_reads += z[index];
			}
		}
		
		while (!done) {
			double prev_ll = ll;
			double[] prev_probs = new double[this.edge_list.length];
			for (int i=0; i<this.edge_list.length; i++) {
				prev_probs[i] = this.edge_list[i].weight;
			}
			
			//mStep();
			mStep(Math.round(num_reads));
			ll = eStep(sub_paths);
			wgt_update = getWeightUpdate(prev_probs);
			if (v == 1) {
				emOutput(ll,iter);
			}
			
			if ((ll-prev_ll) > min_inc_ll && wgt_update > min_edge_up) {
				iter++;
			}
			else {
				iter++;
				if (ll-prev_ll < 0 && Math.abs(ll-prev_ll) > min_inc_ll) {
					System.out.println("ERROR: decreasing likelihood");
					System.err.println("  Number of reads: "+reads.length);
					System.err.println("  Iterations: "+iter);
				}
				if (v == 1) {
					System.out.println();
					System.out.println("Current log likelihood: "+ll);
					System.out.println("Converged after "+iter+" iterations");
					System.out.println("Log likelihood diff: "+(ll-prev_ll));
					System.out.println("Max edge difference: "+wgt_update);
				}
				else {
				    //emOutput(ll,iter);
				}
				done = true;
			}
		}
		return bic(reads,ll);
    }
    
    // Master method for running the EM algorithm
    // Returns the Bayesian Information Criterion value for this splice graph
    public double learnEM(String[] reads, ArrayList[] sub_paths, int v) {
		// convergence conditions
		double min_edge_up = .00001, min_inc_ll = .001;
		boolean done = false;
		
		// things to check
		double ll = eStep(sub_paths);
		int iter = 0;
		double wgt_update = 0;
		
		float num_reads = 0;
		for (int i=0; i<this.edges[0].length; i++) {
			if (this.edges[0][i] != null) {
				int index = getEdgeIndex(this.edges[0][i]);
				num_reads += z[index];
			}
		}
		
		while (!done) {
			double prev_ll = ll;
			double[] prev_probs = new double[this.edge_list.length];
			for (int i=0; i<this.edge_list.length; i++) {
				prev_probs[i] = this.edge_list[i].weight;
			}
			
			//mStep();
			mStep(Math.round(num_reads));
			ll = eStep(sub_paths);
			wgt_update = getWeightUpdate(prev_probs);
			if (v == 1) {
				emOutput(ll,iter);
			}
			
			if ((ll-prev_ll) > min_inc_ll && wgt_update > min_edge_up) {
				iter++;
			}
			else {
				iter++;
				if (ll-prev_ll < 0 && Math.abs(ll-prev_ll) > min_inc_ll) {
					System.err.println("ERROR: decreasing likelihood");
					System.err.println("  Number of reads: "+reads.length);
					System.err.println("  Iterations: "+iter);
				}
				if (v == 1) {
					System.out.println();
					System.out.println("Current log likelihood: "+ll);
					System.out.println("Converged after "+iter+" iterations");
					System.out.println("Log likelihood diff: "+(ll-prev_ll));
					System.out.println("Max edge difference: "+wgt_update);
				}
				else {
				    //emOutput(ll,iter);
				}
				done = true;
			}
		}
		return bic(reads,ll);
    }
	
    // Master method for running the EM algorithm on single bowtie-aligned stuff
    // Returns the Bayesian Information Criterion value for this splice graph
    public double learnEM(String[] reads, ArrayList[] sub_paths,
						  ArrayList[] quality, int v) {
		// convergence conditions
		double min_edge_up = .00001, min_inc_ll = .001;
		boolean done = false;
		
		// things to check
		double ll = eStep(sub_paths,quality);
		int iter = 0;
		double wgt_update = 0;
		if (v==1) {
			emOutput(ll,iter-1);
		}
		
		float num_reads = 0;
		for (int i=0; i<this.edges[0].length; i++) {
			if (this.edges[0][i] != null) {
				int index = getEdgeIndex(this.edges[0][i]);
				num_reads += z[index];
			}
		}
		
		while (!done) {
			double prev_ll = ll;
			double[] prev_probs = new double[this.edge_list.length];
			for (int i=0; i<this.edge_list.length; i++) {
				prev_probs[i] = this.edge_list[i].weight;
			}
			
			//mStep();
			mStep(Math.round(num_reads));
			ll = eStep(sub_paths,quality);
			wgt_update = getWeightUpdate(prev_probs);
			if (v == 1) {
				emOutput(ll,iter);
			}
			
			if ((ll-prev_ll) > min_inc_ll && wgt_update > min_edge_up) {
				iter++;
			}
			else {
				iter++;
				if (ll-prev_ll < 0 && Math.abs(ll-prev_ll) > min_inc_ll) {
					System.err.println("ERROR: decreasing likelihood");
					System.err.println("  Number of reads: "+reads.length);
					System.err.println("  Iterations: "+iter);
				}
				if (v == 1) {
					System.out.println();
					System.out.println("Current log likelihood: "+ll);
					System.out.println("Converged after "+iter+" iterations");
					System.out.println("Log likelihood diff: "+(ll-prev_ll));
					System.out.println("Max edge difference: "+wgt_update);
				}
				else {
				    //emOutput(ll,iter);
				}
				done = true;
			}
		}
		return bic(reads,ll);
    }
    
	// Overloaded method for read-segment alignments
	public ArrayList[] readSubPaths(String[][] reads) {
		ArrayList[] read_list = new ArrayList[reads.length];
		int num_multiples = 0;
		int num_none = 0;
		
		for (int i=0; i<reads.length; i++) {
			read_list[i] = new ArrayList<Integer[]>();
			ArrayList<Integer[]> r1 = alignRead(reads[i][0]);
			ArrayList<Integer[]> r2 = alignRead(reads[i][1]);
			ArrayList<Integer[]> matches = find_paths(r1,r2);
			for (int k=0; k<matches.size(); k++) {
				read_list[i].add(matches.get(k));
			}
			if (matches.size() == 0) { 
				num_none++; 
			}
			else if (matches.size() > 1) { num_multiples++; }
			matches = new ArrayList<Integer[]>();
		}
		//System.out.println("No matches(p): "+num_none);
		return read_list;
	}
	
	// Determines the order of alignments, returns a code
	// -1 : r1 before r2, intersection
	//  1 : r2 before r1, intersection
	//  0 : r1 = r2
	// -2 : r1 before r2, no intersection
	//  2 : r2 before r1, no intersection
	// 10 : INCOMPATIBLE
	// --> Assumes line graph PSG
	public int getOrder(Integer[] r1, Integer[] r2) {
		int retval = 0;
		if (!compAlign(r1,r2)) {
			retval = 10;
		}
		else if (r1[0].intValue() < r2[0].intValue()) {
			// r1 is before r2
			for (int i=0; i<r1.length; i++) {
				for (int j=0; j<r2.length; j++) {
					if (r1[i].equals(r2[j])) { 
						retval = -1;
						break;
					}
				}
				if (retval == -1) { break; }
			}
			if (retval != -1) { retval = -2; }
		}
		else if (r1[0].intValue() > r2[0].intValue()) {
			// r2 is before r1
			for (int i=0; i<r1.length; i++) {
				for (int j=0; j<r2.length; j++) {
					if (r1[i].equals(r2[j])) { 
						retval = 1; 
						break;
					}
				}
				if (retval == 1) { break; }
			}
			if (retval != 1) { retval = 2; }
		}
		else {
			if (compAlign(r1,r2)) { retval = 0; }
			else { retval = 10; }
		}
		return retval;
	}
	
	// checks for compatibility of alignments
	private boolean compAlign(Integer[] a, Integer[] b) {
		boolean retval = true;
		
		// 1. check for overlap
		boolean overlap = false;
		int[] where = new int[2];
		for (int i=0; i<a.length; i++) {
			for (int j=0; j<b.length; j++) {
				if (a[i] == b[j]) {
					// exon in common
					overlap = true;
					where[0] = i;
					where[1] = j;
					break;
				}
			}
			if (overlap) { break; }
		}
		if (overlap) {
			int a_ind = where[0];
			int b_ind = where[1];
			boolean interrupted = false;
			while (a_ind < a.length && b_ind < b.length) {
				if (a[a_ind] != b[b_ind]) {
					interrupted = true;
				}
				a_ind++;
				b_ind++;
			}
			if (interrupted) {
				return false;
			}
			return true;
		}
		
		// 2. no overlap: check for path
		if (a[0] < b[0]) { bfs_paths(a,b); }
		else { bfs_paths(b,a); }
		if (global_paths.size() == 0) { return false; }
		return true;
	}
	
	// Finds all possible paths between paired alignments
	public ArrayList<Integer[]> find_paths(ArrayList<Integer[]> r1,
										   ArrayList<Integer[]> r2) {
		if (r1.size() == 0 || r2.size() == 0) {
			return null;
		}
		ArrayList<Integer[]> paths = new ArrayList<Integer[]>();
		for (int i=0; i<r1.size(); i++) {
			Integer[] p1 = r1.get(i);
			for (int j=0; j<r2.size(); j++) {
				Integer[] p2 = r2.get(j);
				Integer[] path = null;
				int order = getOrder(p1, p2);
				int index;
				ArrayList<Integer> temp = new ArrayList<Integer>();
				switch (order) {
						// case 0 : they start on the same exon
						// can just use the p2-comes-first-with-intersection code
					case 0:
						// case 1 : exon in common, p2 is first
					case 1: 
						for (int k=0; k<p2.length; k++) {
							temp.add(p2[k]);
						}
						index = 0;
						while (index < p1.length && temp.contains(p1[index])) {
							index++;
						}
						while (index < p1.length) {
							temp.add(p1[index]);
							index++;
						}
						path = Utility.IntegerArray(temp);
						if (!Utility.Acontains(paths, path)) {
							paths.add(path);
						}
						break;
						// case -1: exon in common, r1 is first
					case -1:
						for (int k=0; k<p1.length; k++) {
							temp.add(p1[k]);
						}
						index = 0;
						while (index < p2.length && temp.contains(p2[index])) {
							index++;
						}
						while (index < p2.length) {
							temp.add(p2[index]);
							index++;
						}
						path = Utility.IntegerArray(temp);
						if (!Utility.Acontains(paths, path)) {
							paths.add(path);
						}
						break;
						// case 2 : no exons in common, r2 is first
						// may have multiple paths
					case 2: 
						bfs_paths(p2,p1);
						for (int x=0; x<global_paths.size(); x++) {
							for (int k=0; k<p2.length; k++) {
								temp.add(p2[k]);
							}
							index = 0;
							Integer[] p = global_paths.get(x);
							while (index < p.length && temp.contains(p[index])) {
								index++;
							}
							while (index < p.length) {
								temp.add(p[index]);
								index++;
							}
							index = 0;
							while (index < p1.length && temp.contains(p1[index])) {
								index++;
							}
							while (index < p1.length) {
								temp.add(p1[index]);
								index++;
							}
							path = Utility.IntegerArray(temp);
							if (!Utility.Acontains(paths,path)) {
								paths.add(path);
							}
							temp = new ArrayList<Integer>();
						}
						break;
						// case -2: no exons in common, r1 is first
						// may have multiple paths
					case -2: 
						bfs_paths(p1,p2);
						for (int x=0; x<global_paths.size(); x++) {
							for (int k=0; k<p1.length; k++) {
								temp.add(p1[k]);
							}
							index = 0;
							Integer[] p = global_paths.get(x);
							while (index < p.length && temp.contains(p[index])) {
								index++;
							}
							while (index < p.length) {
								temp.add(p[index]);
								index++;
							}
							index = 0;
							while (index < p2.length && temp.contains(p2[index])) {
								index++;
							}
							while (index < p2.length) {
								temp.add(p2[index]);
								index++;
							}
							path = Utility.IntegerArray(temp);
							if (!Utility.Acontains(paths,path)) {
								paths.add(path);
							}
							temp = new ArrayList<Integer>();
						}
						break;
					case 10:
						// bad things
						break;
				}
			}
		}
		return paths;
	}
	
	// Finds all possible paths between paired alignments
	// q1 = quality scores for the paths in r1
	// q2 = quality scores for the paths in r2
	// IMPORTANT: assume all reads have same length
	public ArrayList[] find_paths(ArrayList<Integer[]> r1, ArrayList<Integer[]> r2,
								  ArrayList<Double> q1, ArrayList<Double> q2,
								  ArrayList<Integer> pos1, ArrayList<Integer> pos2,
								  int read_len) {
		if (r1.size() == 0 || r2.size() == 0) {
			return null;
		}
		ArrayList[] paths = new ArrayList[3];
		paths[0] = new ArrayList<Integer[]>();		// alignment paths
		paths[1] = new ArrayList<Double>();			// confidence scores
		paths[2] = new ArrayList<Integer[]>();		// path lengths
		
		for (int i=0; i<r1.size(); i++) {
			Integer[] a = r1.get(i); // first read alignment
			for (int j=0; j<r2.size(); j++) {
				Integer[] b = r2.get(j); // second read alignment
				
				// 1. check for overlap
				boolean overlap = false;
				int[] where = new int[2];
				// iterate through all segments in each alignment,
				// compare pairwise
				for (int k=0; k<a.length; k++) {
					for (int l=0; l<b.length; l++) {
						if (a[k].intValue() == b[l].intValue()) {
							// exon in common
							overlap = true;
							where[0] = k;
							where[1] = l;
							break;
						}
					}
					if (overlap) { break; }
				}
				if (overlap) {
					// important: do not write overlapping segment twice!
					int a_ind = where[0];
					int b_ind = where[1];
					if (a_ind != 0 && b_ind != 0) {
						// case 1: incompatible prefixes before intersecting
					}
					else {
						// add the prefix
						ArrayList<Integer> storage = new ArrayList<Integer>();
						if (a_ind > 0) {
							for (int k=0; k<a_ind; k++) {
								storage.add(a[k]);
							}
						}
						else if (b_ind > 0) {
							for (int k=0; k<b_ind; k++) {
								storage.add(b[k]);
							}
						}
						// case 2: compatible prefix; check suffix
						while (a_ind < a.length && b_ind < b.length &&
							   a[a_ind].intValue() == b[b_ind].intValue()) {
							storage.add(a[a_ind]); // just picking one
							a_ind++;
							b_ind++;
						}
						if (a_ind != a.length && b_ind != b.length) {
							// early termination, incompatible
						}
						else {
							while (a_ind != a.length) {
								storage.add(a[a_ind]);
								a_ind++;
							}
							while (b_ind != b.length) {
								storage.add(b[b_ind]);
								b_ind++;
							}
							paths[0].add(Utility.IntegerArray(storage));
							paths[1].add(q1.get(i)*q2.get(j));
							paths[2].add(getlen(read_len,Utility.IntegerArray(storage),a,b,pos1.get(i),pos2.get(j)));
						}
					}
				}
				else {
					// 2. no overlap: check for path
					if (a[0].intValue() < b[0].intValue()) { bfs_paths(a,b); }
					else { bfs_paths(b,a); }
					for (int k=0; k<global_paths.size(); k++) {
						paths[0].add(global_paths.get(k));
						paths[1].add(q1.get(i)*q2.get(j));
						paths[2].add(getlen(read_len,global_paths.get(k),a,b,pos1.get(i),pos2.get(j)));
					}
				}
			}
		}
		return paths;
	}
	
	private ArrayList<Integer[]> global_paths;
	
	// Breadth first search for all possible paths between reads
	// p1 is aligned before p2
	private void bfs_paths(Integer[] p1, Integer[] p2) {
		this.global_paths = new ArrayList<Integer[]>();
		int startseg = p1[p1.length-1].intValue();
		int goalseg = p2[0].intValue();
		Stack<Integer> stack = new Stack<Integer>();
		ArrayList<Integer> path = new ArrayList<Integer>();
		stack.push(p1[p1.length-1]);
		for (Integer i : p1) {
			path.add(i);
		}
		for (int i=0; i<this.edges[startseg].length; i++) {
			if (this.edges[startseg][i] != null) {
				stack.push(new Integer(i));
			}
		}
		
		while (!stack.empty()) {
			Integer stacktop = stack.peek();
			if (stacktop.intValue() > goalseg) {
				// missed goal : pop
				stack.pop();
			}
			else if (stacktop.intValue() < goalseg) {
				if (path.contains(stacktop)) {
					// been here already : pop
					stack.pop();
					path.remove(path.size()-1);
				}
				else {
					// not at goal yet : keep going
					path.add(stacktop);
					for (int i=0; i<this.edges[stacktop.intValue()].length; i++) {
						if (this.edges[stacktop.intValue()][i] != null) {
							stack.push(new Integer(i));
						}
					}
				}
			}
			else {
				// at goal : add path then pop
				//path.add(stacktop);
				for (Integer i : p2) {
					path.add(i);
				}
				global_paths.add(Utility.IntegerArray(path));
				stack.pop();
				for (int i=0; i<p2.length; i++) {
					path.remove(path.size()-1);
				}
			}
		}
	}
	
	// Preprocessing step: find read-segment alignments
	public ArrayList[] readSubPaths(String[] reads) {
		ArrayList[] read_list = new ArrayList[reads.length];
		int num_multiples = 0;
		int num_none = 0;
		int num_plus2 = 0;
		
		for (int i=0; i<reads.length; i++) {
			read_list[i] = new ArrayList<Integer[]>();
			ArrayList<Integer[]> matches = alignRead(reads[i]);
			for (int j=0; j<matches.size(); j++) {
				read_list[i].add(matches.get(j));
				if (matches.get(j).length > 2) { num_plus2++; }
			}
			if (matches.size() == 0) { num_none++; }
			else if (matches.size() > 1) { num_multiples++; }
			matches = new ArrayList<Integer[]>();
		}
		return read_list;
	}
	
	// Returns segments to which this read aligns
	private ArrayList<Integer[]> alignRead(String read) {
		ArrayList<Integer[]> matches = new ArrayList<Integer[]>();
		int read_len = read.length();
		for (int j=0; j<this.segments.length-1; j++) {
			if (this.segments[j].length() > 0) {
				
				// SINGLE SEGMENT ALIGNMENT
				String compare = this.segments[j].sequence;
				int index = -1;
				do {
					index = compare.indexOf(read);
					if (index > -1) {
						Integer[] match = {j};
						matches.add(match);
						compare = compare.substring(index+1);
					}
				} while (index > -1);
				
				// MULTIPLE SEGMENT ALIGNMENT
				Stack<Segment> stack = new Stack<Segment>();
				int total = 0;
				ArrayList<Integer> alignment = new ArrayList<Integer>();
				stack.push(this.segments[j]); // added to stack
				alignment.add(j);             // marked
				int start = Math.max(0,this.segments[j].length()-(read_len-1));
				compare = this.segments[j].sequence.substring(start);
				int initial_len = compare.length();
				// initialize stack
				for (int k=0; k<this.edges[j].length; k++) {
					if (this.edges[j][k] != null) {
						stack.push(this.segments[k]);
					}
				}
				
				while (!stack.empty()) {
					Integer stacktop = getSegIndex(stack.peek());
					if (alignment.contains(stacktop)) {
						// marked: pop
						int cut = Math.max(compare.length()-stack.peek().length(), 0);
						compare = compare.substring(0,cut);
						total = compare.length()-initial_len;
						stack.pop();
						alignment.remove(alignment.size()-1);
					}
					else if (stack.peek().length()>=(read_len-total-1)
							 || stacktop.intValue()==this.segments.length-1) {
						// long enough: compare then pop
						alignment.add(stacktop);
						int end = (read_len-total-1);
						int cut = compare.length();
						if (stacktop.intValue() < this.segments.length-1) {
							compare += stack.peek().sequence.substring(0,end);
						}
						else {
							for (int k=0; k<read_len-total-1; k++) {
								compare += "A";
							}
						}
						if (compare.indexOf(read) > -1) {
							matches.add(Utility.IntegerArray(alignment));
						}
						compare = compare.substring(0,cut);
						stack.pop();
						alignment.remove(alignment.size()-1);
					}
					else {
						// not long enough: add seq, push adjacent segments
						alignment.add(stacktop);
						int ind = stacktop.intValue();
						int cut = compare.length();
						compare += stack.peek().sequence;
						if (compare.indexOf(read) > -1) {
							matches.add(Utility.IntegerArray(alignment));
							compare = compare.substring(0,cut);
							stack.pop();
							alignment.remove(alignment.size()-1);
						}
						else {
							total += stack.peek().length();
							for (int k=0; k<this.edges[ind].length; k++) {
								if (this.edges[ind][k] != null) {
									stack.push(this.segments[k]);
								}
							}
						}
					}
				}
			}
		}	
		// check to see if we're completely in the poly-A tail
		boolean poly_a = true;
		for (int k=0; k<read_len; k++) {
			if (read.charAt(k) != 'A') { 
				poly_a = false; 
				break;
			}
		}
		if (poly_a /*&& matches.size() == 0*/) {
			// we're in the poly-a tail, just stop here
			Integer[] a = new Integer[1];
			a[0] = new Integer(this.segments.length-1);
			matches.add(a);
		}
		
		return matches;
	}
	
	// Calculates length of a particular fragment based on its subpath
	// s1 is an index into a[0]
	// s2 is an index into b[0]
	private int getlen(int read_len, Integer[] s, Integer[] a, Integer[] b, 
					   Integer s1, Integer s2) {
		int num_bases=0;
		int temp_used = read_len;
		boolean overlap = false;
		int[] where = new int[2];
		// iterate through all segments in each alignment,
		// compare pairwise
		for (int k=0; k<a.length; k++) {
			for (int l=0; l<b.length; l++) {
				if (a[k].intValue() == b[l].intValue()) {
					// exon in common
					overlap = true;
					where[0] = k;
					where[1] = l;
					break;
				}
			}
			if (overlap) { break; }
		}
		if (overlap) {
			// this is easy
			if (where[0] > 0) {
				// a comes first
				num_bases += getSegment(s[0].intValue()).length()-s1;
				num_bases += s2+read_len;
				for (int i=1; !s[i].equals(b[0]); i++) {
					num_bases += getSegment(s[i].intValue()).length();
				}
			}
			else if (where[0] == where[1]) {
				// start in the same segment
				if (s1.intValue() < s2.intValue()) {
					// a comes first
					num_bases = (s2+read_len)-s1;
				}
				else {
					// b comes first (or they're the same, but that's unlikely)
					num_bases = (s1+read_len)-s2;
				}
			}
			else {
				// b comes first
				num_bases += getSegment(s[0].intValue()).length()-s2;
				num_bases += s1+read_len;
				for (int i=1; !s[i].equals(a[0]); i++) {
					num_bases += getSegment(s[i].intValue()).length();
				}
			}
		}
		else {
			// at this point, s1 and s2 are guaranteed to be indexes into the segment
			if (a[0].intValue() == s[0].intValue()) {
				// a comes first
				num_bases += getSegment(s[0].intValue()).length()-s1;
				num_bases += s2+read_len;
				for (int i=1; !s[i].equals(b[0]); i++) {
					// add any intermediate lengths
					num_bases += getSegment(s[i].intValue()).length();
				}
			}
			else {
				// b comes first
				num_bases += getSegment(s[0].intValue()).length()-s2;
				num_bases += s1+read_len;
				for (int i=1; !s[i].equals(a[0]); i++) {
					// add any intermediate lengths
					num_bases += getSegment(s[i].intValue()).length();
				}
			}
			//num_bases = FRAG_MEAN;
		}
		return num_bases;
	}
	
	// Calculates length of a particular fragment based on its subpath/reads
	private int getlen(int read_len, Integer[] s, Integer[] p1, Integer[] p2) {
		int pos1 = p1[0];
		int pos2 = p2[0];
		int end1 = p1[1];
		int end2 = p2[1];
		int start = Math.min(pos1,pos2);
		int end = Math.max(end1,end2);
		int length = 0;
		
		boolean poly_a_0 = (p1[0]==p1[1])?true:false;
		boolean poly_a_1 = (p2[0]==p2[1])?true:false;
		
		if (s.length == 1) {
			length = end-start;
			if (length <= 0) {
				length = FRAG_MEAN;
			}
		}
		else {
			length += this.segments[s[0]].end-start;
			if (!poly_a_0 && !poly_a_1) {
				length += end-this.segments[s[s.length-1]].start;
			}
			else {
				if (poly_a_0) { length += read_len; }
				if (poly_a_1) { length += read_len; }
			}
			for (int i=1; i<s.length-1; i++) {
				length += this.segments[s[i]].length();
			}
		}
		//System.out.println(length);
		return length;
	}
	
	// Calculates length of a particular fragment subpath alignment
	// String[] read: a pair of reads
	private int getlen(String[] read, Integer[] s) {
		int length = 0;
		boolean poly_a_0 = true;
		boolean poly_a_1 = true;
		for (int i=0; i<read[0].length(); i++) {
			if (read[0].charAt(i) != 'A') {
				poly_a_0 = false;
				break;
			}
		}
		for (int i=0; i<read[1].length(); i++) {
			if (read[1].charAt(i) != 'A') {
				poly_a_1 = false;
				break;
			}
		}
		String align = "";
		for (int i=0; i<s.length; i++) {
			align += this.segments[s[i].intValue()].sequence;
		}
		if (s[s.length-1].intValue() == this.segments.length-1) {
			for (int i=0; i<read[0].length(); i++) {
				align += "A";
			}
		}
		int pos1 = align.length();
		int pos2 = align.length();
		if (!poly_a_0) {
			pos1 = align.indexOf(read[0]);
		}
		if (!poly_a_1) {
			pos2 = align.indexOf(read[1]);
		}
		int start = Math.min(pos1,pos2);
		if (start == -1) {
			// this should not happen
			length = -1;
		}
		else {
			align = align.substring(start);
			if (!poly_a_0 && !poly_a_1) {
				int end = Math.max(pos1-start,pos2-start);
				end += read[0].length();
				try {
					align = align.substring(0,end);
				}
				catch (StringIndexOutOfBoundsException siob) {
					System.out.println("Reads: "+read[0]+","+read[1]);
					System.out.println("Align has "+align.length()+" characters, trying to access character "+end);
					System.out.println("start: "+start);
					System.out.println("pos1: "+pos1+", pos2: "+pos2);
					System.exit(0);
				}
				length = align.length();
			}
			else {
				length = align.length()+read[0].length();
			}
		}
		return length;
	}
	
	// Calculates length of a particular fragment based on its subpath/reads
	/*private int getlen(String[] read, Integer[] s) {
	 boolean poly_a_0 = true;
	 boolean poly_a_1 = true;
	 for (int i=0; i<read[0].length(); i++) {
	 if (read[0].charAt(i) != 'A') {
	 poly_a_0 = false;
	 break;
	 }
	 }
	 for (int i=0; i<read[1].length(); i++) {
	 if (read[1].charAt(i) != 'A') {
	 poly_a_1 = false;
	 break;
	 }
	 }
	 int pos1 = getStartPos(read[0],s);
	 if (pos1 == -1) { 
	 System.out.println(read[0]); 
	 }
	 int pos2 = getStartPos(read[1],s);
	 if (pos2 == -1) { 
	 System.out.println(read[1]);
	 }
	 int end1 = getEndPos(read[0],s);
	 int end2 = getEndPos(read[1],s);
	 int start = Math.min(pos1,pos2);
	 int end = Math.max(end1,end2);
	 int length = 0;
	 if (s.length == 1) {
	 length = end-start;
	 if (length <= 0) {
	 length = FRAG_MEAN;
	 }
	 }
	 else {
	 length += this.segments[s[0]].end-start;
	 if (!poly_a_0 && !poly_a_1) {
	 length += end-this.segments[s[s.length-1]].start;
	 }
	 else {
	 if (poly_a_0) { length += read[0].length(); }
	 if (poly_a_1) { length += read[1].length(); }
	 }
	 for (int i=1; i<s.length-1; i++) {
	 length += this.segments[s[i]].length();
	 }
	 }
	 //System.out.println(length);
	 return length;
	 }*/
	
	// Calculates the probability of a given fragment length
	private double p_length(int length) {
		double coeff = 1/Math.sqrt(2*Math.PI*Math.pow(FRAG_STDEV,2));
		double exp = -Math.pow((length-FRAG_MEAN),2)/(double)(2*Math.pow(FRAG_STDEV,2));
		return coeff*Math.exp(exp);
		
		//return 0.001; // uniform [0,1000]
	}
	
	// Calculates the CDF of a given fragment length
	private double cdf_length(double length) {
		double val = length-FRAG_MEAN;
		val /= Math.sqrt(2*Math.pow(FRAG_STDEV,2));
		return .5*(1+erf(val));
	}
	
    // Approximates erf(x)
    private double erf(double x) {
		double a = (8.0*(Math.PI-3.0))/(3.0*Math.PI*(4.0-Math.PI));
		double sgn = (x>=0)?1.0:-1.0;
		double result = Math.exp(-Math.pow(x,2)*(((4.0/Math.PI)+a*Math.pow(x,2))/(1+a*Math.pow(x,2))));
		result = Math.sqrt(1-result);
		result *= sgn;
		return result;
    }
	
    // Expectation for paired reads
    private double eStep(ArrayList[] sub_paths, int[][] path_lengths) {
		calculate_f();
		double log_likelihood = 0.0;
		double avg_len = avg_suff(0);
		double[] edgeprob = new double[edge_list.length];
		z = new double[edge_list.length];
		
		// loop over all paired reads
		for (int i=0; i<sub_paths.length; i++) {
			if (sub_paths[i].size() > 0) {
				double prob = 0;
				// loop over all matches
				for (int j=0; j<sub_paths[i].size(); j++) {
					Integer[] path = (Integer[]) sub_paths[i].get(j);
					int fraglen = path_lengths[i][j];
					double fragprob = 0;
					if (path[path.length-1] == this.segments.length-1) {
						fragprob = 1-(cdf_length(fraglen+.5)+cdf_length(fraglen-.5))/2.0;
					}
					else {
						fragprob = p_length(fraglen);
					}
					if (fragprob > 0) {
						double result = g(path)*fragprob;
						prob += result;
						for (int k=0; k<edge_list.length; k++) {
							edgeprob[k] += g(path,k)*fragprob;
						}
						
					}
				}
				for (int k=0; k<z.length; k++) {
					z[k] += edgeprob[k]/prob;
					edgeprob[k] = 0;
				}
				log_likelihood += Math.log(prob/avg_len);
				prob = 0;
			}
		}
		// add a pseudoread to all edges from nodes with outdegree > 1
		// deprecated; mStep now incorporates priors
		/*
		 for (int i=0; i<z.length; i++) {
		 if (get_outdegree(edge_list[i]) > 1) {
		 Integer[] path = {getSegIndex(edge_list[i].source),getSegIndex(edge_list[i].target)};
		 double prob = g(path);
		 for (int k=0; k<edge_list.length; k++) {
		 edgeprob[k] += g(path,k);
		 }
		 for (int j=0; j<z.length; j++) {
		 z[j] += edgeprob[j]/prob;
		 edgeprob[j] = 0;
		 }
		 log_likelihood += Math.log(prob/avg_len);
		 }
		 }
		 // */
		return log_likelihood;
    }
	
    // Expectation for single, bowtie reads
    private double eStep(ArrayList[] sub_paths, ArrayList[] quality) {
		calculate_f();
		double log_likelihood = 0.0;
		double avg_len = d_avg_suff()[0];
		double[] edgeprob = new double[edge_list.length];
		double[] lengths = new double[edge_list.length];
		z = new double[edge_list.length];
		
		for (int i=0; i<sub_paths.length; i++) {
			if (sub_paths[i].size() > 0) {
				double prob = 0;
				// loop over all matches
				for (int j=0; j<sub_paths[i].size(); j++) {
					Integer[] path = (Integer[]) sub_paths[i].get(j);
					Double q = (Double) quality[i].get(j);
					if (q > 0) {
						// get g(s)
						prob += g(path)*q;
						for (int k=0; k<edge_list.length; k++) {
							// get g(s,i,j) where i=source and j=target for e_ij
							edgeprob[k] += g(path,k)*q;
						}
					}
				}
				if (prob > 0) {
					for (int j=0; j<z.length; j++) {
						z[j] += edgeprob[j]/prob;
						edgeprob[j] = 0;
					}
					log_likelihood += Math.log(prob/avg_len);
				}
				prob = 0;
			}
		}
		// add a pseudoread to all edges from nodes with outdegree > 1
		// deprecated; mStep now incorporates priors
		/*
		 for (int i=0; i<z.length; i++) {
		 if (get_outdegree(edge_list[i]) > 1) {
		 Integer[] path = {getSegIndex(edge_list[i].source),getSegIndex(edge_list[i].target)};
		 double prob = g(path);
		 for (int k=0; k<edge_list.length; k++) {
		 edgeprob[k] += g(path,k);
		 }
		 for (int j=0; j<z.length; j++) {
		 z[j] += edgeprob[j]/prob;
		 edgeprob[j] = 0;
		 }
		 log_likelihood += Math.log(prob/avg_len);
		 }
		 }
		 //*/
		return log_likelihood;
    }
	
    // Expectation for paired reads
    private double eStep(ArrayList[] sub_paths, int[][] path_lengths, ArrayList[] quality) {
		calculate_f();
		double log_likelihood = 0.0;
		double avg_len = d_avg_suff()[0];
		double[] edgeprob = new double[edge_list.length];
		z = new double[edge_list.length];
		
		int num_reads = 0;
		
		// loop over all paired reads
		for (int i=0; i<sub_paths.length; i++) {
			boolean used = false;
			if (sub_paths[i].size() > 0) {
				double prob = 0;
				// loop over all matches
				for (int j=0; j<sub_paths[i].size(); j++) {
					Integer[] path = (Integer[]) sub_paths[i].get(j);
					double q = ((Double) quality[i].get(j)).doubleValue();
					//System.err.println("\t"+q);
					int fraglen = path_lengths[i][j];
					//System.err.println("\t"+fraglen);
					double fragprob = 0;
					if (fraglen < FRAG_MEAN+20*FRAG_STDEV) {
						if (path[path.length-1].intValue() == this.segments.length-1) {
							fragprob = 1-(cdf_length(fraglen+.5)+cdf_length(fraglen-.5))/2.0;
						}
						else {
							fragprob = p_length(fraglen);
						}
						fragprob *= q;
					}
					//System.err.println("\t"+fragprob);
					if (fragprob > 0) {
						used = true;
						double result = g(path)*fragprob;
						//System.err.println("\t"+g(path));
						prob += result;
						for (int k=0; k<edge_list.length; k++) {
							edgeprob[k] += g(path,k)*fragprob;
							//System.err.println(k+": "+edgeprob[k]);
						}
					}
				}
				if (prob > 0) {
					for (int k=0; k<z.length; k++) {
						z[k] += edgeprob[k]/prob;
						edgeprob[k] = 0;
					}
				}
				if (prob > 0) {
					log_likelihood += Math.log(prob/avg_len);
				}
				//System.err.println(i+":"+log_likelihood);
				prob = 0;
			}
			if (used) { num_reads++; }
			//System.err.println(log_likelihood);
		}
		// add a pseudoread to all edges from nodes with outdegree > 1
		// deprecated; mStep now incorporates priors
		/*
		 for (int i=0; i<z.length; i++) {
		 if (get_outdegree(edge_list[i]) > 1) {
		 Integer[] path = {getSegIndex(edge_list[i].source),getSegIndex(edge_list[i].target)};
		 double prob = g(path);
		 for (int k=0; k<edge_list.length; k++) {
		 edgeprob[k] += g(path,k);
		 }
		 for (int j=0; j<z.length; j++) {
		 z[j] += edgeprob[j]/prob;
		 edgeprob[j] = 0;
		 }
		 log_likelihood += Math.log(prob/avg_len);
		 }
		 }
		 //*/
		//print_z();
		//System.err.println("\tNumber of reads used: "+num_reads);
		return log_likelihood;
    }
    
    // Expectation: assign data based on current probabilities
    private double eStep(ArrayList[] sub_paths) {
		calculate_f();
		
		double log_likelihood = 0.0;
		double total_prob = 0.0;
		double avg_len = d_avg_suff()[0];
		double[] edgeprob = new double[edge_list.length];
		double[] lengths = new double[edge_list.length];
		z = new double[edge_list.length];
		
		//System.out.println("g(s):");
		// loop over all reads
		for (int i=0; i<sub_paths.length; i++) {
			if (sub_paths[i].size() > 0) {
				double prob = 0;
				// loop over all matches
				for (int j=0; j<sub_paths[i].size(); j++) {
					// get g(s)
					prob += g((Integer[]) sub_paths[i].get(j));
					for (int k=0; k<edge_list.length; k++) {
						// get g(s,i,j) where i=source and j=target for e_ij
						edgeprob[k] += g((Integer[]) sub_paths[i].get(j),k);
					}
				}
				for (int j=0; j<z.length; j++) {
					z[j] += edgeprob[j]/prob;
					edgeprob[j] = 0;
				}
				log_likelihood += Math.log(prob/avg_len);
				//System.out.printf("\t%d: %.2f ",i,prob);
				prob = 0;
			}
		}
		// add a pseudoread to all edges from nodes with outdegree > 1
		// deprecated; mStep now incorporates priors
		/*
		 for (int i=0; i<z.length; i++) {
		 if (get_outdegree(edge_list[i]) > 1) {
		 Integer[] path = {getSegIndex(edge_list[i].source),getSegIndex(edge_list[i].target)};
		 double prob = g(path);
		 for (int k=0; k<edge_list.length; k++) {
		 edgeprob[k] += g(path,k);
		 }
		 for (int j=0; j<z.length; j++) {
		 z[j] += edgeprob[j]/prob;
		 edgeprob[j] = 0;
		 }
		 log_likelihood += Math.log(prob/avg_len);
		 }
		 }
		 // */
		//System.out.println();
		//print_z();
		return log_likelihood;
    }
    
    // Debug method
    private void print_z() {
		System.err.println("=== Z values ===");
		for (int i=0; i<z.length; i++) {
			System.err.println("\t"+z[i]);
		}
		System.err.println();
    }
    
    // Calculates g(s')
    private double g(Integer[] map) {
		double h = this.f[0][map[0].intValue()];
		if (map.length > 1) {
			for (int j=1; j<map.length; j++) {
				h *= edges[map[j-1].intValue()][map[j].intValue()].weight;
			}
		}
		return h;
    }
    
    // Calculates g(s',i,j)
    private double g(Integer[] map, int edge) {
		int source = getSegIndex(edge_list[edge].source);
		int target = getSegIndex(edge_list[edge].target);
		//System.out.print("("+source+","+target+"): ");
		if (map.length > 1) {
			// check if edge is contained in the path
			for (int i=1; i<map.length; i++) {
				if (map[i-1].intValue() == source && map[i].intValue() == target) {
					//System.out.print("*");
					return g(map);
				}
			}
		}
		double h = edge_list[edge].weight;
		//System.out.print(h+",");
		h *= Math.max(this.f[0][map[0].intValue()]*this.f[map[map.length-1].intValue()][source],
					  this.f[0][source]*this.f[target][map[0].intValue()]);
		//System.out.print(h+",");
		if (map.length > 1) {
			for (int j=1; j<map.length; j++) {
				h *= edges[map[j-1].intValue()][map[j].intValue()].weight;
			}
		}
		//System.out.println(h);
		return h;
    }
	
    // Maximization step, figure out new probabilities based on E-step
	// TODO: Update with Bayesian priors
	private void mStep() {
		//print_z();
		double update, min_update = .001;
		double[] old_wgts = new double[edge_list.length];
		/*for (int i=0; i<this.segments.length; i++) {
		 avg_pr[i] = avg_pref(i);
		 avg_su[i] = avg_suff(i);
		 }*/
		do {
			calculate_f();
			double[] avg_pr = d_avg_pref();
			double[] avg_su = d_avg_suff();
			double[] denoms = new double[segments.length];
			double[] numers = new double[edge_list.length];
			for (int i=0; i<this.edge_list.length; i++) {
				Edge e = this.edge_list[i];
				int source = getSegIndex(e.source);
				int target = getSegIndex(e.target);
				numers[i] = z[i]/(avg_pr[source]+avg_su[target]);
				denoms[source] += numers[i];
			}
			for (int i=0; i<this.edge_list.length; i++) {
				Edge e = this.edge_list[i];
				int source = getSegIndex(e.source);
				old_wgts[i] = e.weight;
				if (denoms[source] > 0) {
					e.weight = numers[i]/denoms[source];
				}
				else {
					// updated logic: do not update
					// e.weight = 0;
				}
			}
			update = getWeightUpdate(old_wgts);
		} while (update > min_update);
    }
    
	// Maximization step with bonus priors
    private void mStep(int num_reads) {
		//print_z();
		double update = 0, min_update = .001;
		double l_update = 0;
		double[] old_wgts = new double[edge_list.length];
		double[] old_lamb = new double[lambda.length];
		boolean[] execute = new boolean[segments.length];
		do {
			double[] avg_pr = d_avg_pref();
			double[] avg_su = d_avg_suff();
			for (int i=0; i<execute.length; i++) {
				execute[i] = true;
			}
			double[] denoms = new double[edge_list.length];
			double[] numers = new double[edge_list.length];
			
			double[] alpha_bar = new double[edge_list.length];
			// find \bar{alpha_ij} = c_ij + b_ij/(N/D(a)f(i)(d_p(i)+d_q(j))+lambda_i)
			for (int i=0; i<this.edge_list.length; i++) {
				Edge e = this.edge_list[i];
				int source = getSegIndex(e.source);
				int target = getSegIndex(e.target);
				numers[i] = z[i] + beta[i];
				denoms[i] = (num_reads/avg_su[0])*f[0][source]*(avg_pr[source]+avg_su[target])+lambda[source];
				alpha_bar[i] = numers[i]/denoms[i];
				if (denoms[i] <= 0 || Double.isNaN(denoms[i]) /*|| numers[i] == 0*/) {
					execute[source] = false;
				}
			}
			//*
			for (int i=0; i<this.edges.length; i++) {
				double total_numerator = 0.0;
				for (int j=0; j<this.edges[i].length; j++) {
					if (this.edges[i][j] != null) {
						total_numerator += numers[getEdgeIndex(edges[i][j])];
					}
				}
				if (total_numerator <= 0.0) {
					execute[i] = false;
 				}
				//System.err.println("seg "+i+": "+total_numerator);
			}
			// */
			// find alpha_ij = \bar{alpha_ij}/sum_j alpha_ijt
			for (int i=0; i<this.edge_list.length; i++) {
				Edge e = this.edge_list[i];
				int source = getSegIndex(e.source);
				double denominator = 0;
				for (int j=0; j<edges[source].length; j++) {
					if (edges[source][j] != null) {
						denominator += alpha_bar[getEdgeIndex(edges[source][j])];
					}
				}
				old_wgts[i] = e.weight;
				if (execute[source] /*&& !Double.isNaN(denominator)*/) {
					//System.err.println("denominator: "+denominator);
					e.weight = alpha_bar[i]/denominator;
				}
				else {
					// updated logic: do not update
					// e.weight = 0;
				}
			}
			update = getWeightUpdate(old_wgts);
			// update lambda_i = c_i + b_i - N/D(a)f(i)(d_p(i)+d_q(j)-ell(i))
			calculate_f();
			for (int i=0; i<lambda.length; i++) {
				double c_i = 0.0;
				double b_i = 0.0;
				for (int j=0; j<edges[i].length; j++) {
					if (edges[i][j] != null) {
						int index = getEdgeIndex(edges[i][j]);
						c_i += z[index];
						b_i += beta[index];
					}
				}
				old_lamb[i] = lambda[i];
				lambda[i] = c_i + b_i - (num_reads/avg_su[0])*f[0][i]*(avg_pr[i]+avg_su[i]-segments[i].length());
			}
			l_update = getLambdaUpdate(old_lamb);
		} while (update > min_update || l_update > min_update);
    }
	
	// Dynamic programming implementation of avg_suff
	private double[] d_avg_suff() {
		double[] suff = new double[this.segments.length];
		for (int i=this.segments.length-1; i>=0; i--) {
			if (i == this.segments.length-1) {
				suff[i] = 0;
			}
			else {
				for (int j=0; j<this.edges.length; j++) {
					if (edges[i][j] != null) {
						suff[i] += edges[i][j].weight*suff[j];
					}
				}
				suff[i] += segments[i].length();
			}
		}
		return suff;
	}
	
	// Dynamic programming implementation of avg_pref
	private double[] d_avg_pref() {
		double[] pref = new double[this.segments.length];
		for (int i=0; i<segments.length; i++) {
			if (i == 0) {
				pref[i] = 0;
			}
			else {
				for (int j=0; j<this.edges.length; j++) {
					if (edges[j][i] != null) {
						pref[i] += f[0][j]*this.edges[j][i].weight*pref[j];
					}
				}
				pref[i] /= f[0][i];
				pref[i] += this.segments[i].length();
			}
		}
		return pref;
	}
	
	// Calculates average prefix length before a given segment
	private double avg_pref(int seg) {
		double l = 0;
		if (seg == 0) {
			// do nothing
		}
		else {
			for (int i=0; i<edges.length; i++) {
				if (edges[i][seg] != null) {
					Edge e = edges[i][seg];
					int source = getSegIndex(e.source);
					l += f[0][source]*e.weight*avg_pref(source);
				}
			}
			l /= f[0][seg];
			l += segments[seg].length();
		}
		return l;
	}
	
	// Calculates average suffix length after a given segment
	private double avg_suff(int seg) {
		double l = 0;
		if (seg == segments.length-1) {
			// do nothing
		}
		else {
			for (int i=0; i<edges.length; i++) {
				if (edges[seg][i] != null) {
					Edge e = edges[seg][i];
					int target = getSegIndex(e.target);
					l += e.weight*avg_suff(target);
				}
			}
			l += segments[seg].length();
		}
		return l;
	}
	
	// Calculates the maximum update amount of the new weights
	// Trusts that oldWgts.length == this.edge_list.length
	private double getWeightUpdate(double[] oldWgts) {
		double update = 0;
		for (int i=0; i<oldWgts.length; i++) {
			double diff = Math.abs(oldWgts[i]-edge_list[i].weight);
			if (diff > update) {
				update = diff;
			}
		}
		return update;
	}
	
	private double getLambdaUpdate(double[] oldWgts) {
		double update = 0;
		for (int i=0; i<oldWgts.length; i++) {
			double diff = Math.abs(oldWgts[i]-lambda[i]);
			if (diff > update) {
				update = diff;
			}
		}
		return update;
	}
	
	// Formatted output for each EM step
	public void emOutput(double ll, int iter) {
		System.out.print((iter+1)+" ");
		System.out.printf("%.3f :: ",ll);
		for (int i=0; i<this.edge_list.length; i++) {
			System.out.printf("%.3f ",edge_list[i].weight);
		}
		System.out.println();
	}
	
	///////////////////////////////////////////////////////////////////////////
	// Structure Modification Methods
	///////////////////////////////////////////////////////////////////////////
	
	// Locates the next segment with multiple incoming edges
	public int nextDup(int seg) {
		int nextSeg = this.segments.length;
		for (int i=seg+1; i<this.segments.length-1; i++) {
			int segCount = getNumInc(i);
			if (segCount > 1) {
				nextSeg = i;
				break;
			}
		}
		return nextSeg;
	}
	
	
	// Primary segment duplication method
	public void duplicateSegment(int seg) {
		int numInc = getNumInc(seg);
		addSegment(seg,numInc);
		int numEdges = addEdges(seg,numInc);
		updateEdgeList(numEdges);
	}
	
	// duplicates a segment and saves the new segment array to this.segments
	private void addSegment(int seg, int numInc) {
		Segment[] newSeg = new Segment[this.segments.length+numInc-1];
		for (int i=0; i<seg; i++) {
			newSeg[i] = this.segments[i];
		}
		for (int i=0; i<numInc; i++) {
			newSeg[i+seg] = new Segment(this.segments[seg]);
		}
		for (int i=seg+1; i<this.segments.length; i++) {
			newSeg[i+numInc-1] = this.segments[i];
		}
		this.segments = newSeg;
	}
	
	// creates new edge matrix for an added segment
	private int addEdges(int seg, int numInc) {
		Edge[][] e = new Edge[this.segments.length][this.segments.length];
		int segInd = seg, numEdges = 0;
		for (int i=0; i<seg; i++) {
			if (this.edges[i][seg] != null) {
				e[i][segInd] = this.edges[i][seg];
				numEdges++;
				e[i][segInd].target = this.segments[segInd];
				segInd++;
			}
		}
		for (int i=seg; i<this.edges.length; i++) {
			if (this.edges[seg][i] != null) {
				for (int j=seg; j<segInd; j++) {
					int target = getSegIndex(this.edges[seg][i].target);
					e[j][target] = new Edge(this.edges[seg][i]);
					numEdges++;
					e[j][target].source = this.segments[j];
				}
			}
		}
		for (int i=0; i<this.edge_list.length; i++) {
			int source = getSegIndex(this.edge_list[i].source);
			int target = getSegIndex(this.edge_list[i].target);
			if ((source < seg || source > seg+numInc-1) && source != -1) {
				if ((target < seg || target > seg+numInc-1) && target != -1) {
					e[source][target] = this.edge_list[i];
					numEdges++;
				}
			}
		}
		this.edges = e;
		return numEdges;
	}
	
	// creates new edge list from current connectivity matrix
	private void updateEdgeList(int numEdges) {
		Edge[] e = new Edge[numEdges];
		int ind = 0;
		for (int i=0; i<this.edges.length; i++) {
			for (int j=0; j<this.edges[i].length; j++) {
				if (this.edges[i][j] != null) {
					e[ind] = this.edges[i][j];
					ind++;
				}
			}
		}
		this.edge_list = e;
	}
	
	///////////////////////////////////////////////////////////////////////////
	// Splice Graph Utilities
	///////////////////////////////////////////////////////////////////////////
	
	// Sets fragment length mean
	public void setMean(int mean) {
		FRAG_MEAN = mean;
	}
	
	// Sets fragment length standard deviation
	public void setStdev(double stdev) {
		FRAG_STDEV = stdev;
	}
	
	// Sets uniform beta value
	public void setBeta(double beta) {
		BETA_VALUE = beta;
		System.out.println("beta = "+BETA_VALUE);
		initialize_lambda(); // should really do this again once we've changed beta...
	}
	
	// Counts number of unaligned reads (HAXX)
	private int unaligned(String[] reads, ArrayList[] a) {
		int num = 0;
		for (int i=0; i<reads.length; i++) {
			if (a[i].size() > 0) {
				num++;
			}
		}
		return num;
	}
	
	// Dynamic programming version of calculate_f()
	private void calculate_f() {
		this.f = new double[this.segments.length][this.segments.length];
		// initial conditions
		for (int i=0; i<this.segments.length; i++) {
			this.f[i][i] = 1;
		}
		
		for (int i=this.segments.length-1; i>=0; i--) {
			for (int j=i+1; j<this.segments.length; j++) {
				if (this.edges[i][j] != null) {
					// direct path
					f[i][j] = this.edges[i][j].weight;
				}
				for (int k=i+1; k<j; k++) {
					if (this.edges[k][j] != null) {
						f[i][j] += f[i][k]*this.edges[k][j].weight;
					}
				}
			}
		}
	}
	
	// debug method!
	private void print_fs() {
		for (int i=0; i<f.length; i++) {
			for (int j=0; j<f[i].length; j++) {
				System.out.print(f[i][j]+" ");
			}
			System.out.println();
		}
	}
	
    // Converts a segment into an index value for this.segments
    public int getSegIndex(Segment s) {
		for (int i=0; i<this.segments.length; i++) {
			if (this.segments[i] == s) {
				return i;
			}
		}
		return -1;
    }
    
	// Converts an edge into an index value for this.edge_list
	public int getEdgeIndex(Edge e) {
		for (int i=0; i<this.edge_list.length; i++) {
			if (this.edge_list[i] == e) {
				return i;
			}
		}
		return -1;
	}
	
	// Sets edge probabilities so that all outgoing edges have equal weight
	public void resetEdgeWeights() {
		int num_out = 0;
		for (int i=0; i<this.segments.length; i++) {
			for (int j=0; j<this.segments.length; j++) {
				if (this.edges[i][j] != null) {
					num_out++;
				}
			}
			for (int j=0; j<this.segments.length; j++) {
				if (this.edges[i][j] != null) {
					this.edges[i][j].weight = 1/(double)num_out;
				}
			}
			num_out = 0;
		}
	}
	
	// Initializes our lagrangian multipliers
	// For now, assuming that we're going to do an outdegree prior?
	public void initialize_lambda() {
		for (int i=0; i<beta.length; i++) {
			beta[i] = BETA_VALUE;
		}
		for (int i=0; i<lambda.length; i++) {
			lambda[i] = (double) get_outdegree(segments[i])*BETA_VALUE;
			//System.out.println("lambda["+i+"] = "+lambda[i]);
		}
	}
	
	// Returns current values of edge weights
	public double[] getEdgeWeights() {
		double[] output = new double[this.edge_list.length];
		for (int i=0; i<this.edge_list.length; i++) {
			output[i] = this.edge_list[i].weight;
		}
		return output;
	}
	
	// Gets log likelihood when you know the alignment already
	public double getLogLikelihood(ArrayList[] sub_paths, String[] reads) {
		return eStep(sub_paths);
	}
	
	// Gets log likelihood for single bowtie-aligned reads
	public double getLogLikelihood(String[] reads, ArrayList[] sub_paths,
								   ArrayList[] quality) {
		return eStep(sub_paths,quality);
	}
	
	// Gets log likelihood for paired reads
	public double getLogLikelihood(String[][] reads) {
		ArrayList[] sub_paths = readSubPaths(reads);
		int[][] path_lengths = new int[sub_paths.length][];
		for (int i=0; i<sub_paths.length; i++) {
			path_lengths[i] = new int[sub_paths[i].size()];
			for (int j=0; j<sub_paths[i].size(); j++) {
				Integer[] path = (Integer[]) sub_paths[i].get(j);
				path_lengths[i][j] = getlen(reads[i],path);
			}
		}
		return eStep(sub_paths,path_lengths);
	}
	
	// Gets log likelihood for paired reads
	public double getLogLikelihood(String[][] reads, ArrayList[] sub_paths) {
		int[][] path_lengths = new int[sub_paths.length][];
		for (int i=0; i<sub_paths.length; i++) {
			path_lengths[i] = new int[sub_paths[i].size()];
			for (int j=0; j<sub_paths[i].size(); j++) {
				Integer[] path = (Integer[]) sub_paths[i].get(j);
				path_lengths[i][j] = getlen(reads[i],path);
			}
		}
		return eStep(sub_paths,path_lengths);
	}
	
	// Gets log likelihood for paired bowtie-aligned reads
	public double getLogLikelihood(String[][] reads, ArrayList[] sub_paths, 
								   ArrayList[] quality, int[][] path_lengths) {
		return eStep(sub_paths,path_lengths,quality);
	}
	
	// Checks the log likelihood of a set of reads against the current weights
	public double getLogLikelihood(String[] reads) {
		ArrayList[] sub_paths = readSubPaths(reads);
		return eStep(sub_paths);
	}
	
	// Prints current splice graph
	public void printSpliceGraph() {
		String filename = "generated.psg";
		try {
			BufferedWriter out = new BufferedWriter(new FileWriter(filename));
			/*out.write("[sequence]");
			 out.newLine();
			 out.write(this.sequence);
			 out.newLine();*/
			out.write("[segments]");
			out.newLine();
			for (int i=0; i<this.segments.length; i++) {
				//out.write(i+"\t"+this.segments[i].start+"\t"+this.segments[i].end);
				out.write(i+"\t"+this.segments[i].sequence);
				out.newLine();
			}
			out.write("[edges]");
			out.newLine();
			for (int i=0; i<this.edge_list.length; i++) {
				out.write(getSegIndex(this.edge_list[i].source)+
						  "\t"+getSegIndex(this.edge_list[i].target)+"\t"+
						  this.edge_list[i].weight);
				out.newLine();
			}
			out.close();
		}
		catch (IOException ioe) {
			System.exit(0);
		}
	}
	
	// Prints current splice graph, allows you to specify filename
	public void printSpliceGraph(String filename) {
		try {
			BufferedWriter out = new BufferedWriter(new FileWriter(filename));
			/*out.write("[sequence]");
			 out.newLine();
			 out.write(this.sequence);
			 out.newLine();*/
			out.write("[segments]");
			out.newLine();
			for (int i=0; i<this.segments.length; i++) {
				//out.write(i+"\t"+this.segments[i].start+"\t"+this.segments[i].end);
				out.write(i+"\t"+this.segments[i].sequence);
				out.newLine();
			}
			out.write("[edges]");
			out.newLine();
			for (int i=0; i<this.edge_list.length; i++) {
				out.write(getSegIndex(this.edge_list[i].source)+
						  "\t"+getSegIndex(this.edge_list[i].target)+"\t"+
						  this.edge_list[i].weight);
				out.newLine();
			}
			out.close();
		}
		catch (IOException ioe) {
			System.exit(0);
		}
	}
	
	// Computes the Bayesian Information Criterion for the current
	// splice graph and reads
	private double bic(String[] reads, double ll) {
		// BIC = ln p(D|\theta_ML) - 0.5 M ln N
		// M = (\sum_v outdegree(v)) - |V| + 1
		double M = this.edge_list.length;
		double N = reads.length;
		M -= this.segments.length;
		M++;
		//System.out.println("BIC="+ll+"-0.5*"+M+"*ln("+N+")");
		return (ll-0.5*M*Math.log(N));
	}
	
	// Bayesian Information Criterion for paired reads
	private double bic(String[][] reads, double ll) {
		double M = this.edge_list.length;
		double N = reads.length*reads[0].length;
		M -= this.segments.length;
		M++;
		//System.out.println("BIC="+ll+"-0.5*"+M+"*ln("+N+")");
		return (ll-0.5*M*Math.log(N));
	}
	
	// Returns the number of incoming edges to a segment
	private int getNumInc(int seg) {
		int numInc = 0;
		for (int i=0; i<seg; i++) {
			if (this.edges[i][seg] != null) {
				numInc++;
			}
		}
		return numInc;
	}
	
	// debug utility - prints edge matrix
	public void printEdgeMatrix() {
		for (int i=0; i<this.edges.length; i++) {
			for (int j=0; j<this.edges[i].length; j++) {
				if (this.edges[i][j] != null) {
					System.out.printf("%.2f ",this.edges[i][j].weight);
				}
				else {
					System.out.print("0.00 ");
				}
			}
			System.out.println();
		}
	}
	
	// for resetting the edge weights so that we can generate more reads
	public void setEdgeWeights(double[] weights) {
		if (weights.length == this.edge_list.length) {
			for (int i=0; i<weights.length; i++) {
				this.edge_list[i].weight = weights[i];
			}
		}
		// else do nothing
	}
	
	// for checking whether an edge connects to the start node
	public boolean is_5prime(Edge e) {
		if (e.source.equals(segments[0])) {
			return true;
		}
		return false;
	}
	
	// for determining whether an edge's source has more than one outgoing edge
	public int get_outdegree(Edge e) {
		int outgoing = 0;
		int index = getSegIndex(e.source);
		for (int i=0; i<edges[index].length; i++) {
			if (edges[index][i] != null) {
				outgoing++;
			}
		}
		return outgoing;
	}
	
	// overloaded segment version of the previous method
	public int get_outdegree(Segment s) {
		int outgoing = 0;
		int index = getSegIndex(s);
		for (int i=0; i<edges[index].length; i++) {
			if (edges[index][i] != null) {
				outgoing++;
			}
		}
		return outgoing;
	}
	
	// accessor methods
	public int getNumSegments() {
		return this.segments.length;
	}
	
	/*public String getSequence() {
	 return this.sequence;
	 }*/
	
	public Segment[] getSegments() {
		return this.segments;
	}
	
	public Segment getSegment(int i) {
		return this.segments[i];
	}
	
	public Edge[] getEdgeList() {
		return this.edge_list;
	}
	
	public Edge[][] getEdges() {
		return this.edges;
	}
	
	public Edge getEdge(int i) {
		return this.edge_list[i];
	}
	
	public int[][] getConnected() {
		return this.connected;
	}
	
}

class Segment {
	public int start;
	public int end;
	public String sequence;
	
	public Segment() {
		start = 0;
		end = 0;
		sequence = "";
	}
	
	public Segment(Segment copy) {
		start = copy.start;
		end = copy.end;
		sequence = new String(copy.sequence);
	}
	
	public int length() {
		return sequence.length();
	}
}

class Edge {
	public Segment source;
	public Segment target;
	public double weight;
	
	public Edge() {
		source = null;
		target = null;
		weight = 0.0;
	}
	
	public Edge(Edge copy) {
		source = copy.source;
		target = copy.target;
		weight = copy.weight;
	}
}

class Queue {
	public QueueNode top;
	public QueueNode bottom;
	
	public void enqueue(int i) {
		if (bottom != null) {
			bottom.next = new QueueNode(i);
			bottom = bottom.next;
		}
		else {
			bottom = new QueueNode(i);
			top = bottom;
		}
	}
	
	public int dequeue() {
		int retval = top.value;
		top = top.next;
		if (top == null) {
			bottom = top;
		}
		return retval;
	}
	
	public boolean isEmpty() {
		if (top == null) {
			return true;
		}
		return false;
	}
	
	public void print() {
		if (!isEmpty()) {
			System.out.print("START ");
			QueueNode temp = top;
			do {
				System.out.print(temp.value+" ");
				temp = temp.next;
			} while (temp != null);
			System.out.println("END");
		}
	}
}

class QueueNode {
	public int value;
	public QueueNode next;
	
	public QueueNode(int i) {
		value = i;
	}
}
